# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2007 Oliver Charles
# Copyright (C) 2007, 2010-2011 Lukáš Lalinský
# Copyright (C) 2007-2011, 2020 Philipp Wolfer
# Copyright (C) 2011 Michael Wiencek
# Copyright (C) 2011-2012 Wieland Hoffmann
# Copyright (C) 2013-2015, 2018-2019, 2021 Laurent Monin
# Copyright (C) 2017 Sambhav Kothari
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from picard import log
from picard.coverart.image import CoverArtImage
from picard.coverart.providers.provider import CoverArtProvider


class CoverArtProviderUrlRelationships(CoverArtProvider):

    """Use cover art link and has_cover_art_at MusicBrainz relationships to get
    cover art"""

    NAME = "UrlRelationships"
    TITLE = N_("Allowed Cover Art URLs")

    def queue_images(self):
        self.match_url_relations(('cover art link', 'has_cover_art_at'),
                                 self._queue_from_relationship)
        return CoverArtProvider.FINISHED

    def _queue_from_relationship(self, url):
        log.debug("Found cover art link in URL relationship")
        self.queue_put(CoverArtImage(url))



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2006-2008, 2011-2012, 2014 Lukáš Lalinský
# Copyright (C) 2007 Nikolai Prokoschenko
# Copyright (C) 2008 Gary van der Merwe
# Copyright (C) 2008 Robert Kaye
# Copyright (C) 2008 Will
# Copyright (C) 2008-2010, 2015, 2018-2023 Philipp Wolfer
# Copyright (C) 2009 Carlin Mangar
# Copyright (C) 2009 David Hilton
# Copyright (C) 2011-2012 Chad Wilson
# Copyright (C) 2011-2013, 2015-2017 Wieland Hoffmann
# Copyright (C) 2011-2014 Michael Wiencek
# Copyright (C) 2013-2014, 2017 Sophist-UK
# Copyright (C) 2013-2023 Laurent Monin
# Copyright (C) 2015 Ohm Patel
# Copyright (C) 2015 samithaj
# Copyright (C) 2016 Rahul Raturi
# Copyright (C) 2016 Simon Legner
# Copyright (C) 2016-2017 Sambhav Kothari
# Copyright (C) 2017 Antonio Larrosa
# Copyright (C) 2017 Frederik “Freso” S. Olesen
# Copyright (C) 2018 Kartik Ohri
# Copyright (C) 2018 Vishal Choudhary
# Copyright (C) 2018 virusMac
# Copyright (C) 2018, 2021-2023 Bob Swift
# Copyright (C) 2019 Timur Enikeev
# Copyright (C) 2020-2021 Gabriel Ferreira
# Copyright (C) 2021 Petit Minion
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from collections import OrderedDict
from copy import deepcopy
import datetime
from functools import partial
import itertools
import os.path

from PyQt5 import (
    QtCore,
    QtGui,
    QtWidgets,
)

from picard import (
    PICARD_APP_ID,
    log,
)
from picard.album import Album
from picard.browser import addrelease
from picard.cluster import (
    Cluster,
    FileList,
)
from picard.config import (
    BoolOption,
    FloatOption,
    IntOption,
    Option,
    SettingConfigSection,
    TextOption,
    get_config,
)
from picard.const import PROGRAM_UPDATE_LEVELS
from picard.const.sys import (
    IS_MACOS,
    IS_WIN,
)
from picard.file import File
from picard.formats import supported_formats
from picard.plugin import ExtensionPoint
from picard.script import get_file_naming_script_presets
from picard.track import Track
from picard.util import (
    IgnoreUpdatesContext,
    icontheme,
    iter_files_from_objects,
    iter_unique,
    open_local_path,
    reconnect,
    restore_method,
    thread,
    throttle,
    webbrowser2,
)
from picard.util.cdrom import (
    DISCID_NOT_LOADED_MESSAGE,
    discid,
    get_cdrom_drives,
)

from picard.ui import PreserveGeometry
from picard.ui.aboutdialog import AboutDialog
from picard.ui.coverartbox import CoverArtBox
from picard.ui.filebrowser import FileBrowser
from picard.ui.infodialog import (
    AlbumInfoDialog,
    ClusterInfoDialog,
    FileInfoDialog,
    TrackInfoDialog,
)
from picard.ui.infostatus import InfoStatus
from picard.ui.itemviews import (
    BaseTreeView,
    MainPanel,
)
from picard.ui.logview import (
    HistoryView,
    LogView,
)
from picard.ui.metadatabox import MetadataBox
from picard.ui.newuserdialog import NewUserDialog
from picard.ui.options.dialog import OptionsDialog
from picard.ui.passworddialog import (
    PasswordDialog,
    ProxyDialog,
)
from picard.ui.pluginupdatedialog import PluginUpdatesDialog
from picard.ui.savewarningdialog import SaveWarningDialog
from picard.ui.scripteditor import (
    ScriptEditorDialog,
    ScriptEditorExamples,
)
from picard.ui.searchdialog.album import AlbumSearchDialog
from picard.ui.searchdialog.track import TrackSearchDialog
from picard.ui.statusindicator import DesktopStatusIndicator
from picard.ui.tagsfromfilenames import TagsFromFileNamesDialog
from picard.ui.util import (
    MultiDirsSelectDialog,
    find_starting_directory,
)


ui_init = ExtensionPoint(label='ui_init')


def register_ui_init(function):
    ui_init.register(function.__module__, function)


class MainWindowActions:
    _create_actions = []

    @classmethod
    def add(cls):
        def decorator(fn):
            cls._create_actions.append(fn)
            return fn
        return decorator

    @classmethod
    def create(cls, parent):
        for create_action in cls._create_actions:
            create_action(parent)


class MainWindow(QtWidgets.QMainWindow, PreserveGeometry):

    defaultsize = QtCore.QSize(780, 560)
    selection_updated = QtCore.pyqtSignal(object)
    ready_for_display = QtCore.pyqtSignal()

    options = [
        Option('persist', 'window_state', QtCore.QByteArray()),
        BoolOption('persist', 'window_maximized', False),
        BoolOption('persist', 'view_metadata_view', True),
        BoolOption('persist', 'view_cover_art', True),
        BoolOption('persist', 'view_toolbar', True),
        BoolOption('persist', 'view_file_browser', False),
        TextOption('persist', 'current_directory', ""),
        FloatOption('persist', 'mediaplayer_playback_rate', 1.0),
        IntOption('persist', 'mediaplayer_volume', 50),
    ]

    def __init__(self, parent=None, disable_player=False):
        super().__init__(parent)
        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_NativeWindow)
        self.__shown = False
        app = QtCore.QCoreApplication.instance()
        self._is_wayland = app.is_wayland
        self.selected_objects = []
        self.ignore_selection_changes = IgnoreUpdatesContext(self.update_selection)
        self.toolbar = None
        self.player = None
        self.status_indicators = []
        if DesktopStatusIndicator:
            self.ready_for_display.connect(self._setup_desktop_status_indicator)
        if not disable_player:
            from picard.ui.playertoolbar import Player
            player = Player(self)
            if player.available:
                self.player = player
                self.player.error.connect(self._on_player_error)

        self.script_editor_dialog = None
        self.examples = None

        self.tagger.pluginmanager.updates_available.connect(self.show_plugin_update_dialog)

        self.check_and_repair_naming_scripts()
        self.check_and_repair_profiles()

        self.setupUi()

    def setupUi(self):
        self.setWindowTitle(_("MusicBrainz Picard"))
        icon = QtGui.QIcon()
        for size in (16, 24, 32, 48, 128, 256):
            icon.addFile(
                ":/images/{size}x{size}/{app_id}.png".format(
                    size=size, app_id=PICARD_APP_ID),
                QtCore.QSize(size, size)
            )
        self.setWindowIcon(icon)

        self.show_close_window = IS_MACOS

        self.create_actions()
        self.create_statusbar()
        self.create_toolbar()
        self.create_menus()

        if IS_MACOS:
            self.setUnifiedTitleAndToolBarOnMac(True)

        main_layout = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical)
        main_layout.setObjectName('main_window_bottom_splitter')
        main_layout.setChildrenCollapsible(False)
        main_layout.setContentsMargins(0, 0, 0, 0)

        self.panel = MainPanel(self, main_layout)
        self.panel.setObjectName('main_panel_splitter')
        self.file_browser = FileBrowser(self.panel)
        if not self.show_file_browser_action.isChecked():
            self.file_browser.hide()
        self.panel.insertWidget(0, self.file_browser)

        self.log_dialog = LogView(self)
        self.history_dialog = HistoryView(self)

        self.metadata_box = MetadataBox(self)
        self.cover_art_box = CoverArtBox(self)
        metadata_view_layout = QtWidgets.QHBoxLayout()
        metadata_view_layout.setContentsMargins(0, 0, 0, 0)
        metadata_view_layout.setSpacing(0)
        metadata_view_layout.addWidget(self.metadata_box, 1)
        metadata_view_layout.addWidget(self.cover_art_box, 0)
        self.metadata_view = QtWidgets.QWidget()
        self.metadata_view.setLayout(metadata_view_layout)

        self.show_metadata_view()
        self.show_cover_art()

        main_layout.addWidget(self.panel)
        main_layout.addWidget(self.metadata_view)
        self.setCentralWidget(main_layout)

        # accessibility
        self.set_tab_order()

        for function in ui_init:
            function(self)

    def set_processing(self, processing=True):
        self.panel.set_processing(processing)

    def set_sorting(self, sorting=True):
        self.panel.set_sorting(sorting)

    def keyPressEvent(self, event):
        # On macOS Command+Backspace triggers the so called "Forward Delete".
        # It should be treated the same as the Delete button.
        is_forward_delete = IS_MACOS and \
            event.key() == QtCore.Qt.Key.Key_Backspace and \
            event.modifiers() & QtCore.Qt.KeyboardModifier.ControlModifier
        if event.matches(QtGui.QKeySequence.StandardKey.Delete) or is_forward_delete:
            if self.metadata_box.hasFocus():
                self.metadata_box.remove_selected_tags()
            else:
                self.remove()
        elif event.matches(QtGui.QKeySequence.StandardKey.Find):
            self.search_edit.setFocus(True)
        else:
            super().keyPressEvent(event)

    def show(self):
        self.restoreWindowState()
        super().show()
        self.show_new_user_dialog()
        if self.tagger.autoupdate_enabled:
            self.auto_update_check()
        self.check_for_plugin_update()
        self.metadata_box.restore_state()

    def showEvent(self, event):
        if not self.__shown:
            self.ready_for_display.emit()
            self.__shown = True
        super().showEvent(event)

    def closeEvent(self, event):
        config = get_config()
        if config.setting['quit_confirmation'] and not self.show_quit_confirmation():
            event.ignore()
            return
        if self.player:
            config.persist['mediaplayer_playback_rate'] = self.player.playback_rate()
            config.persist['mediaplayer_volume'] = self.player.volume()
        self.saveWindowState()
        # Confirm loss of unsaved changes in script editor.
        if self.script_editor_dialog:
            if not self.script_editor_dialog.unsaved_changes_in_profile_confirmation():
                event.ignore()
                return
            else:
                # Silently close the script editor without displaying the confirmation a second time.
                self.script_editor_dialog.loading = True
        event.accept()

    def _setup_desktop_status_indicator(self):
        if DesktopStatusIndicator:
            self.register_status_indicator(DesktopStatusIndicator(self.windowHandle()))

    def register_status_indicator(self, indicator):
        self.status_indicators.append(indicator)

    def show_quit_confirmation(self):
        unsaved_files = sum(a.get_num_unsaved_files() for a in self.tagger.albums.values())
        QMessageBox = QtWidgets.QMessageBox

        if unsaved_files > 0:
            msg = QMessageBox(self)
            msg.setIcon(QMessageBox.Icon.Question)
            msg.setWindowModality(QtCore.Qt.WindowModality.WindowModal)
            msg.setWindowTitle(_("Unsaved Changes"))
            msg.setText(_("Are you sure you want to quit Picard?"))
            txt = ngettext(
                "There is %d unsaved file. Closing Picard will lose all unsaved changes.",
                "There are %d unsaved files. Closing Picard will lose all unsaved changes.",
                unsaved_files) % unsaved_files
            msg.setInformativeText(txt)
            cancel = msg.addButton(QMessageBox.StandardButton.Cancel)
            msg.setDefaultButton(cancel)
            msg.addButton(_("&Quit Picard"), QMessageBox.ButtonRole.YesRole)
            ret = msg.exec_()

            if ret == QMessageBox.StandardButton.Cancel:
                return False

        return True

    def saveWindowState(self):
        config = get_config()
        config.persist['window_state'] = self.saveState()
        isMaximized = int(self.windowState()) & QtCore.Qt.WindowState.WindowMaximized != 0
        self.save_geometry()
        config.persist['window_maximized'] = isMaximized
        config.persist['view_metadata_view'] = self.show_metadata_view_action.isChecked()
        config.persist['view_cover_art'] = self.show_cover_art_action.isChecked()
        config.persist['view_toolbar'] = self.show_toolbar_action.isChecked()
        config.persist['view_file_browser'] = self.show_file_browser_action.isChecked()
        self.file_browser.save_state()
        self.panel.save_state()
        self.metadata_box.save_state()

    @restore_method
    def restoreWindowState(self):
        config = get_config()
        self.restoreState(config.persist['window_state'])
        self.restore_geometry()
        if config.persist['window_maximized']:
            self.setWindowState(QtCore.Qt.WindowState.WindowMaximized)
        splitters = config.persist['splitters_MainWindow']
        if splitters is None or 'main_window_bottom_splitter' not in splitters:
            self.centralWidget().setSizes([366, 194])
        self.file_browser.restore_state()

    def create_statusbar(self):
        """Creates a new status bar."""
        self.statusBar().showMessage(_("Ready"))
        infostatus = InfoStatus(self)
        self._progress = infostatus.get_progress
        self.listening_label = QtWidgets.QLabel()
        self.listening_label.setStyleSheet("QLabel { margin: 0 4px 0 4px; }")
        self.listening_label.setVisible(False)
        self.listening_label.setToolTip("<qt/>" + _(
            "Picard listens on this port to integrate with your browser. When "
            "you \"Search\" or \"Open in Browser\" from Picard, clicking the "
            "\"Tagger\" button on the web page loads the release into Picard."
        ))
        self.statusBar().addPermanentWidget(infostatus)
        self.statusBar().addPermanentWidget(self.listening_label)
        self.tagger.tagger_stats_changed.connect(self.update_statusbar_stats)
        self.tagger.listen_port_changed.connect(self.update_statusbar_listen_port)
        self.register_status_indicator(infostatus)
        self.update_statusbar_stats()

    @throttle(100)
    def update_statusbar_stats(self):
        """Updates the status bar information."""
        total_files = len(self.tagger.files)
        total_albums = len(self.tagger.albums)
        pending_files = File.num_pending_files
        pending_requests = self.tagger.webservice.num_pending_web_requests
        for indicator in self.status_indicators:
            indicator.update(files=total_files, albums=total_albums,
                pending_files=pending_files, pending_requests=pending_requests, progress=self._progress())

    def update_statusbar_listen_port(self, listen_port):
        if listen_port:
            self.listening_label.setVisible(True)
            self.listening_label.setText(_("Listening on port %(port)d") % {"port": listen_port})
        else:
            self.listening_label.setVisible(False)

    def set_statusbar_message(self, message, *args, **kwargs):
        """Set the status bar message.

        *args are passed to % operator, if args[0] is a mapping it is used for
        named place holders values
        >>> w.set_statusbar_message("File %(filename)s", {'filename': 'x.txt'})

        Keyword arguments:
        `echo` parameter defaults to `log.debug`, called before message is
        translated, it can be disabled passing None or replaced by ie.
        `log.error`. If None, skipped.

        `translate` is a method called on message before it is sent to history
        log and status bar, it defaults to `_()`. If None, skipped.

        `timeout` defines duration of the display in milliseconds

        `history` is a method called with translated message as argument, it
        defaults to `log.history_info`. If None, skipped.

        Empty messages are never passed to echo and history functions but they
        are sent to status bar (ie. to clear it).
        """
        def isdict(obj):
            return hasattr(obj, 'keys') and hasattr(obj, '__getitem__')

        echo = kwargs.get('echo', log.debug)
        # _ is defined using builtins.__dict__, so setting it as default named argument
        # value doesn't work as expected
        translate = kwargs.get('translate', _)
        timeout = kwargs.get('timeout', 0)
        history = kwargs.get('history', log.history_info)
        if len(args) == 1 and isdict(args[0]):
            # named place holders
            mparms = args[0]
        else:
            # simple place holders, ensure compatibility
            mparms = args
        if message:
            if echo:
                echo(message % mparms)
            if translate:
                message = translate(message)
            message = message % mparms
            if history:
                history(message)
        thread.to_main(self.statusBar().showMessage, message, timeout)

    def _on_submit_acoustid(self):
        if self.tagger.use_acoustid:
            config = get_config()
            if not config.setting['acoustid_apikey']:
                msg = QtWidgets.QMessageBox(self)
                msg.setIcon(QtWidgets.QMessageBox.Icon.Information)
                msg.setWindowModality(QtCore.Qt.WindowModality.WindowModal)
                msg.setWindowTitle(_("AcoustID submission not configured"))
                msg.setText(_(
                    "You need to configure your AcoustID API key before you can submit fingerprints."))
                open_options = QtWidgets.QPushButton(
                    icontheme.lookup('preferences-desktop'), _("Open AcoustID options"))
                msg.addButton(QtWidgets.QMessageBox.StandardButton.Cancel)
                msg.addButton(open_options, QtWidgets.QMessageBox.ButtonRole.YesRole)
                msg.exec_()
                if msg.clickedButton() == open_options:
                    self.show_options('fingerprinting')
            else:
                self.tagger.acoustidmanager.submit()

    @MainWindowActions.add()
    def _create_options_action(self):
        action = QtWidgets.QAction(icontheme.lookup('preferences-desktop'), _("&Options…"), self)
        action.setMenuRole(QtWidgets.QAction.MenuRole.PreferencesRole)
        action.triggered.connect(self.show_options)
        self.options_action = action

    @MainWindowActions.add()
    def _create_show_script_editor_action(self):
        action = QtWidgets.QAction(_("Open &file naming script editor…"))
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+Shift+S")))
        action.triggered.connect(self.open_file_naming_script_editor)
        self.show_script_editor_action = action

    @MainWindowActions.add()
    def _create_cut_action(self):
        action = QtWidgets.QAction(icontheme.lookup('edit-cut', icontheme.ICON_SIZE_MENU), _("&Cut"), self)
        action.setShortcut(QtGui.QKeySequence.StandardKey.Cut)
        action.setEnabled(False)
        action.triggered.connect(self.cut)
        self.cut_action = action

    @MainWindowActions.add()
    def _create_paste_action(self):
        action = QtWidgets.QAction(icontheme.lookup('edit-paste', icontheme.ICON_SIZE_MENU), _("&Paste"), self)
        action.setShortcut(QtGui.QKeySequence.StandardKey.Paste)
        action.setEnabled(False)
        action.triggered.connect(self.paste)
        self.paste_action = action

    @MainWindowActions.add()
    def _create_help_action(self):
        action = QtWidgets.QAction(_("&Help…"), self)
        action.setShortcut(QtGui.QKeySequence.StandardKey.HelpContents)
        action.triggered.connect(self.show_help)
        self.help_action = action

    @MainWindowActions.add()
    def _create_about_action(self):
        action = QtWidgets.QAction(_("&About…"), self)
        action.setMenuRole(QtWidgets.QAction.MenuRole.AboutRole)
        action.triggered.connect(self.show_about)
        self.about_action = action

    @MainWindowActions.add()
    def _create_donate_action(self):
        action = QtWidgets.QAction(_("&Donate…"), self)
        action.triggered.connect(self.open_donation_page)
        self.donate_action = action

    @MainWindowActions.add()
    def _create_report_bug_action(self):
        action = QtWidgets.QAction(_("&Report a Bug…"), self)
        action.triggered.connect(self.open_bug_report)
        self.report_bug_action = action

    @MainWindowActions.add()
    def _create_support_forum_action(self):
        action = QtWidgets.QAction(_("&Support Forum…"), self)
        action.triggered.connect(self.open_support_forum)
        self.support_forum_action = action

    @MainWindowActions.add()
    def _create_add_files_action(self):
        action = QtWidgets.QAction(icontheme.lookup('document-open'), _("&Add Files…"), self)
        action.setStatusTip(_("Add files to the tagger"))
        # TR: Keyboard shortcut for "Add Files…"
        action.setShortcut(QtGui.QKeySequence.StandardKey.Open)
        action.triggered.connect(self.add_files)
        self.add_files_action = action

    @MainWindowActions.add()
    def _create_add_directory_action(self):
        action = QtWidgets.QAction(icontheme.lookup('folder'), _("Add Fold&er…"), self)
        action.setStatusTip(_("Add a folder to the tagger"))
        # TR: Keyboard shortcut for "Add Directory…"
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+E")))
        action.triggered.connect(self.add_directory)
        self.add_directory_action = action

    @MainWindowActions.add()
    def _create_close_window_action(self):
        if self.show_close_window:
            action = QtWidgets.QAction(_("Close Window"), self)
            action.setShortcut(QtGui.QKeySequence(_("Ctrl+W")))
            action.triggered.connect(self.close_active_window)
        else:
            action = None
        self.close_window_action = action

    @MainWindowActions.add()
    def _create_save_action(self):
        action = QtWidgets.QAction(icontheme.lookup('document-save'), _("&Save"), self)
        action.setStatusTip(_("Save selected files"))
        # TR: Keyboard shortcut for "Save"
        action.setShortcut(QtGui.QKeySequence.StandardKey.Save)
        action.setEnabled(False)
        action.triggered.connect(self.save)
        self.save_action = action

    @MainWindowActions.add()
    def _create_submit_acoustid_action(self):
        action = QtWidgets.QAction(icontheme.lookup('acoustid-fingerprinter'), _("S&ubmit AcoustIDs"), self)
        action.setStatusTip(_("Submit acoustic fingerprints"))
        action.setEnabled(False)
        action.triggered.connect(self._on_submit_acoustid)
        self.submit_acoustid_action = action

    @MainWindowActions.add()
    def _create_exit_action(self):
        action = QtWidgets.QAction(_("E&xit"), self)
        action.setMenuRole(QtWidgets.QAction.MenuRole.QuitRole)
        # TR: Keyboard shortcut for "Exit"
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+Q")))
        action.triggered.connect(self.close)
        self.exit_action = action

    @MainWindowActions.add()
    def _create_remove_action(self):
        action = QtWidgets.QAction(icontheme.lookup('list-remove'), _("&Remove"), self)
        action.setStatusTip(_("Remove selected files/albums"))
        action.setEnabled(False)
        action.triggered.connect(self.remove)
        self.remove_action = action

    @MainWindowActions.add()
    def _create_browser_lookup_action(self):
        action = QtWidgets.QAction(icontheme.lookup('lookup-musicbrainz'), _("Lookup in &Browser"), self)
        action.setStatusTip(_("Lookup selected item on MusicBrainz website"))
        action.setEnabled(False)
        # TR: Keyboard shortcut for "Lookup in Browser"
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+Shift+L")))
        action.triggered.connect(self.browser_lookup)
        self.browser_lookup_action = action

    @MainWindowActions.add()
    def _create_submit_cluster_action(self):
        if addrelease.is_available():
            action = QtWidgets.QAction(_("Submit cluster as release…"), self)
            action.setStatusTip(_("Submit cluster as a new release to MusicBrainz"))
            action.setEnabled(False)
            action.triggered.connect(self.submit_cluster)
        else:
            action = None
        self.submit_cluster_action = action

    @MainWindowActions.add()
    def _create_submit_file_as_recording_action(self):
        if addrelease.is_available():
            action = QtWidgets.QAction(_("Submit file as standalone recording…"), self)
            action.setStatusTip(_("Submit file as a new recording to MusicBrainz"))
            action.setEnabled(False)
            action.triggered.connect(self.submit_file)
        else:
            action = None
        self.submit_file_as_recording_action = action

    @MainWindowActions.add()
    def _create_submit_file_as_release_action(self):
        if addrelease.is_available():
            action = QtWidgets.QAction(_("Submit file as release…"), self)
            action.setStatusTip(_("Submit file as a new release to MusicBrainz"))
            action.setEnabled(False)
            action.triggered.connect(partial(self.submit_file, as_release=True))
        else:
            action = None
        self.submit_file_as_release_action = action

    @MainWindowActions.add()
    def _create_similar_items_search_action(self):
        action = QtWidgets.QAction(icontheme.lookup('system-search'), _("Search for similar items…"), self)
        action.setIconText(_("Similar items"))
        action.setStatusTip(_("View similar releases or recordings and optionally choose a different one"))
        action.setEnabled(False)
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+T")))
        action.triggered.connect(self.show_similar_items_search)
        self.similar_items_search_action = action

    @MainWindowActions.add()
    def _create_album_search_action(self):
        action = QtWidgets.QAction(icontheme.lookup('system-search'), _("Search for similar albums…"), self)
        action.setStatusTip(_("View similar releases and optionally choose a different release"))
        action.setEnabled(False)
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+T")))
        action.triggered.connect(self.show_more_albums)
        self.album_search_action = action

    @MainWindowActions.add()
    def _create_track_search_action(self):
        action = QtWidgets.QAction(icontheme.lookup('system-search'), _("Search for similar tracks…"), self)
        action.setStatusTip(_("View similar tracks and optionally choose a different release"))
        action.setEnabled(False)
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+T")))
        action.triggered.connect(self.show_more_tracks)
        self.track_search_action = action

    @MainWindowActions.add()
    def _create_album_other_versions_action(self):
        action = QtWidgets.QAction(_("Show &other album versions…"), self)
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+Shift+O")))
        action.triggered.connect(self.show_album_other_versions)
        self.album_other_versions_action = action

    @MainWindowActions.add()
    def _create_show_file_browser_action(self):
        config = get_config()
        action = QtWidgets.QAction(_("File &Browser"), self)
        action.setCheckable(True)
        if config.persist['view_file_browser']:
            action.setChecked(True)
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+B")))
        action.triggered.connect(self.show_file_browser)
        self.show_file_browser_action = action

    @MainWindowActions.add()
    def _create_show_metadata_view_action(self):
        config = get_config()
        action = QtWidgets.QAction(_("&Metadata"), self)
        action.setCheckable(True)
        if config.persist['view_metadata_view']:
            action.setChecked(True)
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+Shift+M")))
        action.triggered.connect(self.show_metadata_view)
        self.show_metadata_view_action = action

    @MainWindowActions.add()
    def _create_show_cover_art_action(self):
        config = get_config()
        action = QtWidgets.QAction(_("&Cover Art"), self)
        action.setCheckable(True)
        if config.persist['view_cover_art']:
            action.setChecked(True)
        action.setEnabled(self.show_metadata_view_action.isChecked())
        action.triggered.connect(self.show_cover_art)
        self.show_cover_art_action = action

    @MainWindowActions.add()
    def _create_show_toolbar_action(self):
        config = get_config()
        action = QtWidgets.QAction(_("&Actions"), self)
        action.setCheckable(True)
        if config.persist['view_toolbar']:
            action.setChecked(True)
        action.triggered.connect(self.show_toolbar)
        self.show_toolbar_action = action

    @MainWindowActions.add()
    def _create_search_action(self):
        action = QtWidgets.QAction(icontheme.lookup('system-search'), _("Search"), self)
        action.setEnabled(False)
        action.triggered.connect(self.search)
        self.search_action = action

    @MainWindowActions.add()
    def _create_cd_lookup_action(self):
        action = QtWidgets.QAction(icontheme.lookup('media-optical'), _("Lookup &CD…"), self)
        action.setStatusTip(_("Lookup the details of the CD in your drive"))
        # TR: Keyboard shortcut for "Lookup CD"
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+K")))
        action.triggered.connect(self.tagger.lookup_cd)
        self.cd_lookup_action = action

        menu = QtWidgets.QMenu(_("Lookup &CD…"))
        menu.setIcon(icontheme.lookup('media-optical'))
        menu.triggered.connect(self.tagger.lookup_cd)
        self.cd_lookup_menu = menu
        if discid is None:
            log.warning("CDROM: discid library not found - Lookup CD functionality disabled")
            self.cd_lookup_action.setEnabled(False)
            self.cd_lookup_menu.setEnabled(False)
        else:
            thread.run_task(get_cdrom_drives, self._update_cd_lookup_actions)

    @MainWindowActions.add()
    def _create_analyze_action(self):
        action = QtWidgets.QAction(icontheme.lookup('picard-analyze'), _("&Scan"), self)
        action.setStatusTip(_("Use AcoustID audio fingerprint to identify the files by the actual music, even if they have no metadata"))
        action.setEnabled(False)
        action.setToolTip(_("Identify the file using its AcoustID audio fingerprint"))
        # TR: Keyboard shortcut for "Analyze"
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+Y")))
        action.triggered.connect(self.analyze)
        self.analyze_action = action

    @MainWindowActions.add()
    def _create_generate_fingerprints_action(self):
        action = QtWidgets.QAction(icontheme.lookup('fingerprint'), _("&Generate AcoustID Fingerprints"), self)
        action.setIconText(_("Generate Fingerprints"))
        action.setStatusTip(_("Generate the AcoustID audio fingerprints for the selected files without doing a lookup"))
        action.setEnabled(False)
        action.setToolTip(_("Generate the AcoustID audio fingerprints for the selected files"))
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+Shift+Y")))
        action.triggered.connect(self.generate_fingerprints)
        self.generate_fingerprints_action = action

    @MainWindowActions.add()
    def _create_cluster_action(self):
        action = QtWidgets.QAction(icontheme.lookup('picard-cluster'), _("Cl&uster"), self)
        action.setStatusTip(_("Cluster files into album clusters"))
        action.setEnabled(False)
        # TR: Keyboard shortcut for "Cluster"
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+U")))
        action.triggered.connect(self.cluster)
        self.cluster_action = action

    @MainWindowActions.add()
    def _create_autotag_action(self):
        action = QtWidgets.QAction(icontheme.lookup('picard-auto-tag'), _("&Lookup"), self)
        tip = _("Lookup selected items in MusicBrainz")
        action.setToolTip(tip)
        action.setStatusTip(tip)
        action.setEnabled(False)
        # TR: Keyboard shortcut for "Lookup"
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+L")))
        action.triggered.connect(self.autotag)
        self.autotag_action = action

    @MainWindowActions.add()
    def _create_view_info_action(self):
        action = QtWidgets.QAction(icontheme.lookup('picard-edit-tags'), _("&Info…"), self)
        action.setEnabled(False)
        # TR: Keyboard shortcut for "Info"
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+I")))
        action.triggered.connect(self.view_info)
        self.view_info_action = action

    @MainWindowActions.add()
    def _create_refresh_action(self):
        action = QtWidgets.QAction(icontheme.lookup('view-refresh', icontheme.ICON_SIZE_MENU), _("&Refresh"), self)
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+R")))
        action.triggered.connect(self.refresh)
        self.refresh_action = action

    @MainWindowActions.add()
    def _create_enable_renaming_action(self):
        config = get_config()
        action = QtWidgets.QAction(_("&Rename Files"), self)
        action.setCheckable(True)
        action.setChecked(config.setting['rename_files'])
        action.triggered.connect(self.toggle_rename_files)
        self.enable_renaming_action = action

    @MainWindowActions.add()
    def _create_enable_moving_action(self):
        config = get_config()
        action = QtWidgets.QAction(_("&Move Files"), self)
        action.setCheckable(True)
        action.setChecked(config.setting['move_files'])
        action.triggered.connect(self.toggle_move_files)
        self.enable_moving_action = action

    @MainWindowActions.add()
    def _create_enable_tag_saving_action(self):
        config = get_config()
        action = QtWidgets.QAction(_("Save &Tags"), self)
        action.setCheckable(True)
        action.setChecked(not config.setting['dont_write_tags'])
        action.triggered.connect(self.toggle_tag_saving)
        self.enable_tag_saving_action = action

    @MainWindowActions.add()
    def _create_tags_from_filenames_action(self):
        action = QtWidgets.QAction(icontheme.lookup('picard-tags-from-filename'), _("Tags From &File Names…"), self)
        action.setIconText(_("Parse File Names…"))
        action.setToolTip(_("Set tags based on the file names"))
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+Shift+T")))
        action.setEnabled(False)
        action.triggered.connect(self.open_tags_from_filenames)
        self.tags_from_filenames_action = action

    @MainWindowActions.add()
    def _create_open_collection_in_browser_action(self):
        config = get_config()
        action = QtWidgets.QAction(_("&Open My Collections in Browser"), self)
        action.setEnabled(config.setting['username'] != '')
        action.triggered.connect(self.open_collection_in_browser)
        self.open_collection_in_browser_action = action

    @MainWindowActions.add()
    def _create_view_log_action(self):
        action = QtWidgets.QAction(_("View &Error/Debug Log"), self)
        # TR: Keyboard shortcut for "View Error/Debug Log"
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+G")))
        action.triggered.connect(self.show_log)
        self.view_log_action = action

    @MainWindowActions.add()
    def _create_view_history_action(self):
        action = QtWidgets.QAction(_("View Activity &History"), self)
        # TR: Keyboard shortcut for "View Activity History"
        # On macOS ⌘+H is a system shortcut to hide the window. Use ⌘+Shift+H instead.
        action.setShortcut(QtGui.QKeySequence(_("Ctrl+Shift+H") if IS_MACOS else _("Ctrl+H")))
        action.triggered.connect(self.show_history)
        self.view_history_action = action

    @MainWindowActions.add()
    def _create_play_file_action(self):
        action = QtWidgets.QAction(icontheme.lookup('play-music'), _("Open in &Player"), self)
        action.setStatusTip(_("Play the file in your default media player"))
        action.setEnabled(False)
        action.triggered.connect(self.play_file)
        self.play_file_action = action

    @MainWindowActions.add()
    def _create_open_folder_action(self):
        action = QtWidgets.QAction(icontheme.lookup('folder', icontheme.ICON_SIZE_MENU), _("Open Containing &Folder"), self)
        action.setStatusTip(_("Open the containing folder in your file explorer"))
        action.setEnabled(False)
        action.triggered.connect(self.open_folder)
        self.open_folder_action = action

    @MainWindowActions.add()
    def _create_check_update_action(self):
        if self.tagger.autoupdate_enabled:
            action = QtWidgets.QAction(_("&Check for Update…"), self)
            action.setMenuRole(QtWidgets.QAction.MenuRole.ApplicationSpecificRole)
            action.triggered.connect(self.do_update_check)
        else:
            action = None
        self.check_update_action = action

    def create_actions(self):
        MainWindowActions.create(self)

        webservice_manager = self.tagger.webservice.manager
        webservice_manager.authenticationRequired.connect(self.show_password_dialog)
        webservice_manager.proxyAuthenticationRequired.connect(self.show_proxy_dialog)

    def _update_cd_lookup_actions(self, result=None, error=None):
        if error:
            log.error("CDROM: Error on CD-ROM drive detection: %r", error)
        else:
            self.update_cd_lookup_drives(result)

    def update_cd_lookup_drives(self, drives):
        self.cd_lookup_menu.clear()
        self.cd_lookup_action.setEnabled(discid is not None)
        if not drives:
            log.warning(DISCID_NOT_LOADED_MESSAGE)
        else:
            config = get_config()
            shortcut_drive = config.setting['cd_lookup_device'].split(",")[0] if len(drives) > 1 else ""
            for drive in drives:
                action = self.cd_lookup_menu.addAction(drive)
                action.setData(drive)
                if drive == shortcut_drive:
                    self._update_cd_lookup_default_action(action)
        self._set_cd_lookup_from_file_actions(drives)
        self._update_cd_lookup_button()

    def _set_cd_lookup_from_file_actions(self, drives):
        if self.cd_lookup_menu.actions():
            self.cd_lookup_menu.addSeparator()
        action = self.cd_lookup_menu.addAction(_("From CD ripper &log file…"))
        if not drives:
            self._update_cd_lookup_default_action(action)
        action.setData('logfile:eac')

    def _update_cd_lookup_default_action(self, action):
        if action:
            reconnect(self.cd_lookup_action.triggered, action.trigger)
        else:
            reconnect(self.cd_lookup_action.triggered, self.tagger.lookup_cd)

    def _update_cd_lookup_button(self):
        button = self.toolbar.widgetForAction(self.cd_lookup_action)
        enabled = bool(self.cd_lookup_menu.actions() and discid)
        self.cd_lookup_menu.setEnabled(enabled)
        if button:
            if enabled:
                button.setPopupMode(QtWidgets.QToolButton.ToolButtonPopupMode.MenuButtonPopup)
                button.setMenu(self.cd_lookup_menu)
            else:
                button.setMenu(None)

    def toggle_rename_files(self, checked):
        config = get_config()
        config.setting['rename_files'] = checked
        self.update_script_editor_examples()

    def toggle_move_files(self, checked):
        config = get_config()
        config.setting['move_files'] = checked
        self.update_script_editor_examples()

    def toggle_tag_saving(self, checked):
        config = get_config()
        config.setting['dont_write_tags'] = not checked

    def get_selected_or_unmatched_files(self):
        if self.selected_objects:
            files = list(iter_files_from_objects(self.selected_objects))
            if files:
                return files
        return self.tagger.unclustered_files.files

    def open_tags_from_filenames(self):
        files = self.get_selected_or_unmatched_files()
        if files:
            dialog = TagsFromFileNamesDialog(files, self)
            dialog.exec_()

    def open_collection_in_browser(self):
        self.tagger.collection_lookup()

    def create_menus(self):
        menu = self.menuBar().addMenu(_("&File"))
        menu.addAction(self.add_directory_action)
        menu.addAction(self.add_files_action)
        if self.show_close_window:
            menu.addAction(self.close_window_action)
        menu.addSeparator()
        menu.addAction(self.play_file_action)
        menu.addAction(self.open_folder_action)
        menu.addSeparator()
        menu.addAction(self.save_action)
        menu.addAction(self.submit_acoustid_action)
        menu.addSeparator()
        menu.addAction(self.exit_action)
        menu = self.menuBar().addMenu(_("&Edit"))
        menu.addAction(self.cut_action)
        menu.addAction(self.paste_action)
        menu.addSeparator()
        menu.addAction(self.view_info_action)
        menu.addAction(self.remove_action)
        menu = self.menuBar().addMenu(_("&View"))
        menu.addAction(self.show_file_browser_action)
        menu.addAction(self.show_metadata_view_action)
        menu.addAction(self.show_cover_art_action)
        menu.addSeparator()
        menu.addAction(self.show_toolbar_action)
        menu.addAction(self.search_toolbar_toggle_action)
        if self.player:
            menu.addAction(self.player_toolbar_toggle_action)
        menu = self.menuBar().addMenu(_("&Options"))
        menu.addAction(self.enable_renaming_action)
        menu.addAction(self.enable_moving_action)
        menu.addAction(self.enable_tag_saving_action)
        menu.addSeparator()

        self.script_quick_selector_menu = QtWidgets.QMenu(_("&Select file naming script"))
        self.script_quick_selector_menu.setIcon(icontheme.lookup('document-open'))
        self.make_script_selector_menu()

        menu.addMenu(self.script_quick_selector_menu)
        menu.addAction(self.show_script_editor_action)
        menu.addSeparator()

        self.profile_quick_selector_menu = QtWidgets.QMenu(_("&Enable/disable profiles"))
        # self.profile_quick_selector_menu.setIcon(icontheme.lookup('document-open'))
        self.make_profile_selector_menu()

        menu.addMenu(self.profile_quick_selector_menu)
        menu.addSeparator()
        menu.addAction(self.options_action)
        menu = self.menuBar().addMenu(_("&Tools"))
        menu.addAction(self.refresh_action)
        menu.addMenu(self.cd_lookup_menu)
        menu.addAction(self.autotag_action)
        menu.addAction(self.analyze_action)
        menu.addAction(self.cluster_action)
        menu.addAction(self.browser_lookup_action)
        menu.addAction(self.similar_items_search_action)
        menu.addAction(self.album_other_versions_action)
        menu.addSeparator()
        menu.addAction(self.generate_fingerprints_action)
        menu.addAction(self.tags_from_filenames_action)
        menu.addAction(self.open_collection_in_browser_action)
        self.menuBar().addSeparator()
        menu = self.menuBar().addMenu(_("&Help"))
        menu.addAction(self.help_action)
        menu.addSeparator()
        menu.addAction(self.view_history_action)
        menu.addSeparator()
        if self.tagger.autoupdate_enabled:
            menu.addAction(self.check_update_action)
            menu.addSeparator()
        menu.addAction(self.support_forum_action)
        menu.addAction(self.report_bug_action)
        menu.addAction(self.view_log_action)
        menu.addSeparator()
        menu.addAction(self.donate_action)
        menu.addAction(self.about_action)

    def update_toolbar_style(self):
        config = get_config()
        style = QtCore.Qt.ToolButtonStyle.ToolButtonIconOnly
        if config.setting['toolbar_show_labels']:
            style = QtCore.Qt.ToolButtonStyle.ToolButtonTextUnderIcon
        self.toolbar.setToolButtonStyle(style)
        if self.player:
            self.player.toolbar.setToolButtonStyle(style)

    def create_toolbar(self):
        self.create_search_toolbar()
        if self.player:
            self.create_player_toolbar()
        self.create_action_toolbar()
        self.update_toolbar_style()

    def create_action_toolbar(self):
        if self.toolbar:
            self.toolbar.clear()
            self.removeToolBar(self.toolbar)
        self.toolbar = toolbar = QtWidgets.QToolBar(_("Actions"))
        self.insertToolBar(self.search_toolbar, self.toolbar)
        toolbar.setObjectName('main_toolbar')
        if self._is_wayland:
            toolbar.setFloatable(False)  # https://bugreports.qt.io/browse/QTBUG-92191
        if IS_MACOS:
            toolbar.setMovable(False)

        def add_toolbar_action(action):
            toolbar.addAction(action)
            widget = toolbar.widgetForAction(action)
            widget.setFocusPolicy(QtCore.Qt.FocusPolicy.TabFocus)
            widget.setAttribute(QtCore.Qt.WidgetAttribute.WA_MacShowFocusRect)

        config = get_config()
        for action in config.setting['toolbar_layout']:
            if action == 'cd_lookup_action':
                add_toolbar_action(self.cd_lookup_action)
                self._update_cd_lookup_button()
            elif action == 'separator':
                toolbar.addSeparator()
            else:
                try:
                    add_toolbar_action(getattr(self, action))
                except AttributeError:
                    log.warning("Warning: Unknown action name '%r' found in config. Ignored.", action)
        self.show_toolbar()

    def create_player_toolbar(self):
        """"Create a toolbar with internal player control elements"""
        toolbar = self.player.create_toolbar()
        self.addToolBar(QtCore.Qt.ToolBarArea.BottomToolBarArea, toolbar)
        if self._is_wayland:
            toolbar.setFloatable(False)  # https://bugreports.qt.io/browse/QTBUG-92191
        self.player_toolbar_toggle_action = toolbar.toggleViewAction()
        toolbar.hide()  # Hide by default

    def create_search_toolbar(self):
        config = get_config()
        self.search_toolbar = toolbar = self.addToolBar(_("Search"))
        self.search_toolbar_toggle_action = self.search_toolbar.toggleViewAction()
        toolbar.setObjectName('search_toolbar')
        if self._is_wayland:
            toolbar.setFloatable(False)  # https://bugreports.qt.io/browse/QTBUG-92191
        if IS_MACOS:
            self.search_toolbar.setMovable(False)

        search_panel = QtWidgets.QWidget(toolbar)
        hbox = QtWidgets.QHBoxLayout(search_panel)
        self.search_combo = QtWidgets.QComboBox(search_panel)
        self.search_combo.addItem(_("Album"), 'album')
        self.search_combo.addItem(_("Artist"), 'artist')
        self.search_combo.addItem(_("Track"), 'track')
        hbox.addWidget(self.search_combo, 0)
        self.search_edit = QtWidgets.QLineEdit(search_panel)
        self.search_edit.setClearButtonEnabled(True)
        self.search_edit.returnPressed.connect(self.trigger_search_action)
        self.search_edit.textChanged.connect(self.enable_search)
        hbox.addWidget(self.search_edit, 0)
        self.search_button = QtWidgets.QToolButton(search_panel)
        self.search_button.setAutoRaise(True)
        self.search_button.setDefaultAction(self.search_action)
        self.search_button.setIconSize(QtCore.QSize(22, 22))
        self.search_button.setAttribute(QtCore.Qt.WidgetAttribute.WA_MacShowFocusRect)

        # search button contextual menu, shortcut to toggle search options
        def search_button_menu(position):
            menu = QtWidgets.QMenu()
            opts = OrderedDict([
                ('use_adv_search_syntax', N_("&Advanced search")),
                ('builtin_search', N_("&Builtin search"))
            ])

            def toggle_opt(opt, checked):
                config.setting[opt] = checked

            for opt, label in opts.items():
                action = QtWidgets.QAction(_(label), menu)
                action.setCheckable(True)
                action.setChecked(config.setting[opt])
                action.triggered.connect(partial(toggle_opt, opt))
                menu.addAction(action)
            menu.exec_(self.search_button.mapToGlobal(position))

        self.search_button.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.search_button.customContextMenuRequested.connect(search_button_menu)
        hbox.addWidget(self.search_button)
        toolbar.addWidget(search_panel)

    def set_tab_order(self):
        tab_order = self.setTabOrder
        tw = self.toolbar.widgetForAction
        prev_action = None
        current_action = None
        # Setting toolbar widget tab-orders for accessibility
        config = get_config()
        for action in config.setting['toolbar_layout']:
            if action != 'separator':
                try:
                    current_action = tw(getattr(self, action))
                except AttributeError:
                    # No need to log warnings since we have already
                    # done it once in create_toolbar
                    pass

            if prev_action is not None and prev_action != current_action:
                tab_order(prev_action, current_action)

            prev_action = current_action

        tab_order(prev_action, self.search_combo)
        tab_order(self.search_combo, self.search_edit)
        tab_order(self.search_edit, self.search_button)
        # Panels
        tab_order(self.search_button, self.file_browser)
        self.panel.tab_order(tab_order, self.file_browser, self.metadata_box)

    def enable_submit(self, enabled):
        """Enable/disable the 'Submit fingerprints' action."""
        self.submit_acoustid_action.setEnabled(enabled)

    def enable_cluster(self, enabled):
        """Enable/disable the 'Cluster' action."""
        self.cluster_action.setEnabled(enabled)

    def enable_search(self):
        """Enable/disable the 'Search' action."""
        self.search_action.setEnabled(bool(self.search_edit.text()))

    def trigger_search_action(self):
        if self.search_action.isEnabled():
            self.search_action.trigger()

    def search_mbid_found(self, entity, mbid):
        self.search_edit.setText('%s:%s' % (entity, mbid))

    def search(self):
        """Search for album, artist or track on the MusicBrainz website."""
        text = self.search_edit.text()
        entity = self.search_combo.itemData(self.search_combo.currentIndex())
        config = get_config()
        self.tagger.search(text, entity,
                           config.setting['use_adv_search_syntax'],
                           mbid_matched_callback=self.search_mbid_found)

    def add_files(self):
        """Add files to the tagger."""
        current_directory = find_starting_directory()
        formats = []
        extensions = []
        for exts, name in supported_formats():
            exts = ["*" + e.lower() for e in exts]
            if not exts:
                continue
            if not IS_MACOS and not IS_WIN:
                # Also consider upper case extensions
                # macOS and Windows usually support case sensitive file names. Furthermore on both systems
                # the file dialog filters list all extensions we provide, which becomes a bit long when we give the
                # full list twice. Hence only do this trick on other operating systems.
                exts.extend([e.upper() for e in exts])
            exts.sort()
            formats.append("%s (%s)" % (name, " ".join(exts)))
            extensions.extend(exts)
        formats.sort()
        extensions.sort()
        formats.insert(0, _("All supported formats") + " (%s)" % " ".join(extensions))
        formats.insert(1, _("All files") + " (*)")
        files, _filter = QtWidgets.QFileDialog.getOpenFileNames(self, "", current_directory, ";;".join(formats))
        if files:
            config = get_config()
            config.persist['current_directory'] = os.path.dirname(files[0])
            self.tagger.add_files(files)

    def add_directory(self):
        """Add directory to the tagger."""
        current_directory = find_starting_directory()

        dir_list = []
        config = get_config()
        if not config.setting['toolbar_multiselect']:
            directory = QtWidgets.QFileDialog.getExistingDirectory(self, "", current_directory)
            if directory:
                dir_list.append(directory)
        else:
            file_dialog = MultiDirsSelectDialog(self, "", current_directory)
            if file_dialog.exec_() == QtWidgets.QDialog.DialogCode.Accepted:
                dir_list = file_dialog.selectedFiles()

        dir_count = len(dir_list)
        if dir_count:
            parent = os.path.dirname(dir_list[0]) if dir_count > 1 else dir_list[0]
            config.persist['current_directory'] = parent
            if dir_count > 1:
                self.set_statusbar_message(
                    N_("Adding multiple directories from '%(directory)s' …"),
                    {'directory': parent}
                )
            else:
                self.set_statusbar_message(
                    N_("Adding directory: '%(directory)s' …"),
                    {'directory': dir_list[0]}
                )

            self.tagger.add_paths(dir_list)

    def close_active_window(self):
        self.tagger.activeWindow().close()

    def show_about(self):
        return AboutDialog.show_instance(self)

    def show_options(self, page=None):
        options_dialog = OptionsDialog.show_instance(page, self)
        options_dialog.finished.connect(self.options_closed)
        if self.script_editor_dialog is not None:
            # Disable signal processing to avoid saving changes not processed with "Make It So!"
            for signal in self.script_editor_signals:
                signal.disconnect()

        return options_dialog

    def options_closed(self):
        if self.script_editor_dialog is not None:
            self.open_file_naming_script_editor()
            self.script_editor_dialog.show()
        else:
            self.show_script_editor_action.setEnabled(True)
        self.make_profile_selector_menu()
        self.make_script_selector_menu()

    def show_help(self):
        webbrowser2.open('documentation')

    def _show_log_dialog(self, dialog):
        dialog.show()
        dialog.raise_()
        dialog.activateWindow()

    def show_log(self):
        self._show_log_dialog(self.log_dialog)

    def show_history(self):
        self._show_log_dialog(self.history_dialog)

    def open_bug_report(self):
        webbrowser2.open('troubleshooting')

    def open_support_forum(self):
        webbrowser2.open('forum')

    def open_donation_page(self):
        webbrowser2.open('donate')

    def save(self):
        """Tell the tagger to save the selected objects."""
        config = get_config()
        if config.setting['file_save_warning']:
            count = len(self.tagger.get_files_from_objects(self.selected_objects))
            msg = SaveWarningDialog(self, count)
            proceed_with_save, disable_warning = msg.show()
            config.setting['file_save_warning'] = not disable_warning
        else:
            proceed_with_save = True
        if proceed_with_save:
            self.tagger.save(self.selected_objects)

    def remove(self):
        """Tell the tagger to remove the selected objects."""
        self.panel.remove(self.selected_objects)

    def analyze(self):
        def callback(fingerprinting_system):
            if fingerprinting_system:
                self.tagger.analyze(self.selected_objects)
        self._ensure_fingerprinting_configured(callback)

    def generate_fingerprints(self):
        def callback(fingerprinting_system):
            if fingerprinting_system:
                self.tagger.generate_fingerprints(self.selected_objects)
        self._ensure_fingerprinting_configured(callback)

    def play_file(self):
        for file in iter_files_from_objects(self.selected_objects):
            open_local_path(file.filename)

    def _on_player_error(self, error, msg):
        self.set_statusbar_message(msg, echo=log.warning, translate=None)

    def open_folder(self):
        folders = iter_unique(
            os.path.dirname(f.filename) for f
            in iter_files_from_objects(self.selected_objects))
        for folder in folders:
            open_local_path(folder)

    def _ensure_fingerprinting_configured(self, callback):
        config = get_config()

        if not config.setting['fingerprinting_system']:
            if self._show_analyze_settings_info():
                def on_finished(result):
                    callback(config.setting['fingerprinting_system'])

                dialog = self.show_options('fingerprinting')
                dialog.finished.connect(on_finished)
        else:
            callback(config.setting['fingerprinting_system'])

    def _show_analyze_settings_info(self):
        ret = QtWidgets.QMessageBox.question(self,
            _("Configuration Required"),
            _("Audio fingerprinting is not yet configured. Would you like to configure it now?"),
            QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No,
            QtWidgets.QMessageBox.StandardButton.Yes)
        return ret == QtWidgets.QMessageBox.StandardButton.Yes

    def get_first_obj_with_type(self, type):
        for obj in self.selected_objects:
            if isinstance(obj, type):
                return obj
        return None

    def show_similar_items_search(self):
        obj = self.get_first_obj_with_type(Cluster)
        if obj:
            self.show_more_albums()
        else:
            self.show_more_tracks()

    def show_more_tracks(self):
        if not self.selected_objects:
            return
        obj = self.selected_objects[0]
        if isinstance(obj, Track) and obj.files:
            obj = obj.files[0]
        if not isinstance(obj, File):
            log.debug("show_more_tracks expected a File, got %r", obj)
            return
        dialog = TrackSearchDialog(self, force_advanced_search=True)
        dialog.show_similar_tracks(obj)
        dialog.exec_()

    def show_more_albums(self):
        obj = self.get_first_obj_with_type(Cluster)
        if not obj:
            log.debug("show_more_albums expected a Cluster, got %r", obj)
            return
        dialog = AlbumSearchDialog(self, force_advanced_search=True)
        dialog.show_similar_albums(obj)
        dialog.exec_()

    def show_album_other_versions(self):
        obj = self.get_first_obj_with_type(Album)
        if obj and obj.release_group:
            AlbumSearchDialog.show_releasegroup_search(obj.release_group.id, obj)

    def view_info(self, default_tab=0):
        try:
            selected = self.selected_objects[0]
        except IndexError:
            return
        if isinstance(selected, Album):
            dialog_class = AlbumInfoDialog
        elif isinstance(selected, Cluster):
            dialog_class = ClusterInfoDialog
        elif isinstance(selected, Track):
            dialog_class = TrackInfoDialog
        else:
            try:
                selected = next(iter_files_from_objects(self.selected_objects))
            except StopIteration:
                return
            dialog_class = FileInfoDialog
        dialog = dialog_class(selected, self)
        dialog.ui.tabWidget.setCurrentIndex(default_tab)
        dialog.exec_()

    def cluster(self):
        # Cluster all selected unclustered files. If there are no selected
        # unclustered files cluster all unclustered files.
        files = (
            f for f in iter_files_from_objects(self.selected_objects)
            if f.parent == self.tagger.unclustered_files
        )
        try:
            file = next(files)
        except StopIteration:
            files = self.tagger.unclustered_files.files
        else:
            files = itertools.chain([file], files)
        self.tagger.cluster(files, self._cluster_finished)

    def _cluster_finished(self):
        self.panel.update_current_view()
        # Select clusters if no other item or only empty unclustered files item is selected
        if not self.selected_objects or (len(self.selected_objects) == 1
                and self.tagger.unclustered_files in self.selected_objects
                and not self.tagger.unclustered_files.files):
            self.panel.select_object(self.tagger.clusters)
        self.update_actions()

    def refresh(self):
        self.tagger.refresh(self.selected_objects)

    def browser_lookup(self):
        if not self.selected_objects:
            return
        self.tagger.browser_lookup(self.selected_objects[0])

    def submit_cluster(self):
        if self.selected_objects and self._check_add_release():
            for obj in self.selected_objects:
                if isinstance(obj, Cluster):
                    addrelease.submit_cluster(obj)

    def submit_file(self, as_release=False):
        if self.selected_objects and self._check_add_release():
            for file in iter_files_from_objects(self.selected_objects):
                addrelease.submit_file(file, as_release=as_release)

    def _check_add_release(self):
        if addrelease.is_enabled():
            return True
        ret = QtWidgets.QMessageBox.question(self,
            _("Browser integration not enabled"),
            _("Submitting releases to MusicBrainz requires the browser integration to be enabled. Do you want to enable the browser integration now?"),
            QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No,
            QtWidgets.QMessageBox.StandardButton.Yes)
        if ret == QtWidgets.QMessageBox.StandardButton.Yes:
            config = get_config()
            config.setting['browser_integration'] = True
            self.tagger.update_browser_integration()
            if addrelease.is_enabled():
                return True
            else:
                # Something went wrong, let the user configure browser integration manually
                self.show_options('network')
                return False
        else:
            return False

    @throttle(100)
    def update_actions(self):
        can_remove = False
        can_save = False
        can_analyze = False
        can_refresh = False
        can_autotag = False
        can_submit = False
        single = self.selected_objects[0] if len(self.selected_objects) == 1 else None
        can_view_info = bool(single and single.can_view_info())
        can_browser_lookup = bool(single and single.can_browser_lookup())
        is_file = bool(single and isinstance(single, (File, Track)))
        is_album = bool(single and isinstance(single, Album))
        is_cluster = bool(single and isinstance(single, Cluster) and not single.special)

        if not self.selected_objects:
            have_objects = have_files = False
        else:
            have_objects = True
            try:
                next(iter_files_from_objects(self.selected_objects))
                have_files = True
            except StopIteration:
                have_files = False
            for obj in self.selected_objects:
                if obj is None:
                    continue
                # using x = x or obj.x() form prevents calling function
                # if x is already True
                can_analyze = can_analyze or obj.can_analyze()
                can_autotag = can_autotag or obj.can_autotag()
                can_refresh = can_refresh or obj.can_refresh()
                can_remove = can_remove or obj.can_remove()
                can_save = can_save or obj.can_save()
                can_submit = can_submit or obj.can_submit()
                # Skip further loops if all values now True.
                if (
                    can_analyze
                    and can_autotag
                    and can_refresh
                    and can_remove
                    and can_save
                    and can_submit
                ):
                    break
        self.remove_action.setEnabled(can_remove)
        self.save_action.setEnabled(can_save)
        self.view_info_action.setEnabled(can_view_info)
        self.analyze_action.setEnabled(can_analyze)
        self.generate_fingerprints_action.setEnabled(have_files)
        self.refresh_action.setEnabled(can_refresh)
        self.autotag_action.setEnabled(can_autotag)
        self.browser_lookup_action.setEnabled(can_browser_lookup)
        self.play_file_action.setEnabled(have_files)
        self.open_folder_action.setEnabled(have_files)
        self.cut_action.setEnabled(have_objects)
        if self.submit_cluster_action:
            self.submit_cluster_action.setEnabled(can_submit)
        if self.submit_file_as_recording_action:
            self.submit_file_as_recording_action.setEnabled(have_files)
        if self.submit_file_as_release_action:
            self.submit_file_as_release_action.setEnabled(have_files)
        files = self.get_selected_or_unmatched_files()
        self.tags_from_filenames_action.setEnabled(bool(files))
        self.similar_items_search_action.setEnabled(is_file or is_cluster)
        self.track_search_action.setEnabled(is_file)
        self.album_search_action.setEnabled(is_cluster)
        self.album_other_versions_action.setEnabled(is_album)

    def update_selection(self, objects=None, new_selection=True, drop_album_caches=False):
        if self.ignore_selection_changes:
            return

        if objects is not None:
            self.selected_objects = objects
        else:
            objects = self.selected_objects

        self.update_actions()

        obj = None

        # Clear any existing status bar messages
        self.set_statusbar_message("")

        if self.player:
            self.player.set_objects(self.selected_objects)

        metadata_visible = self.metadata_view.isVisible()
        coverart_visible = metadata_visible and self.cover_art_box.isVisible()

        if len(objects) == 1:
            obj = list(objects)[0]
            if isinstance(obj, File):
                if obj.state == obj.ERROR:
                    msg = N_("%(filename)s (error: %(error)s)")
                    mparms = {
                        'filename': obj.filename,
                        'error': obj.errors[0] if obj.errors else ''
                    }
                else:
                    msg = N_("%(filename)s")
                    mparms = {
                        'filename': obj.filename,
                    }
                self.set_statusbar_message(msg, mparms, echo=None, history=None)
            elif isinstance(obj, Track):
                if obj.num_linked_files == 1:
                    file = obj.files[0]
                    if file.has_error():
                        msg = N_("%(filename)s (%(similarity)d%%) (error: %(error)s)")
                        mparms = {
                            'filename': file.filename,
                            'similarity': file.similarity * 100,
                            'error': file.errors[0] if file.errors else ''
                        }
                    else:
                        msg = N_("%(filename)s (%(similarity)d%%)")
                        mparms = {
                            'filename': file.filename,
                            'similarity': file.similarity * 100,
                        }
                    self.set_statusbar_message(msg, mparms, echo=None,
                                               history=None)
        elif coverart_visible and new_selection:
            # Create a temporary file list which allows changing cover art for all selected files
            files = list(iter_files_from_objects(objects))
            obj = FileList(files)

        if coverart_visible and new_selection:
            self.cover_art_box.set_item(obj)

        if metadata_visible:
            if new_selection:
                self.metadata_box.selection_dirty = True
            self.metadata_box.update(drop_album_caches=drop_album_caches)
        self.selection_updated.emit(objects)
        self.update_script_editor_example_files()

    def refresh_metadatabox(self):
        self.tagger.window.metadata_box.selection_dirty = True
        self.tagger.window.metadata_box.update()

    def show_metadata_view(self):
        """Show/hide the metadata view (including the cover art box)."""
        show = self.show_metadata_view_action.isChecked()
        self.metadata_view.setVisible(show)
        self.show_cover_art_action.setEnabled(show)
        if show:
            self.update_selection()

    def show_cover_art(self):
        """Show/hide the cover art box."""
        show = self.show_cover_art_action.isChecked()
        self.cover_art_box.setVisible(show)
        if show:
            self.update_selection()

    def show_toolbar(self):
        """Show/hide the Action toolbar."""
        if self.show_toolbar_action.isChecked():
            self.toolbar.show()
        else:
            self.toolbar.hide()

    def show_file_browser(self):
        """Show/hide the file browser."""
        if self.show_file_browser_action.isChecked():
            sizes = self.panel.sizes()
            if sizes[0] == 0:
                sizes[0] = sum(sizes) // 4
                self.panel.setSizes(sizes)
            self.file_browser.show()
        else:
            self.file_browser.hide()

    def show_password_dialog(self, reply, authenticator):
        config = get_config()
        if reply.url().host() == config.setting['server_host']:
            ret = QtWidgets.QMessageBox.question(self,
                _("Authentication Required"),
                _("Picard needs authorization to access your personal data on the MusicBrainz server. Would you like to log in now?"),
                QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No,
                QtWidgets.QMessageBox.StandardButton.Yes)
            if ret == QtWidgets.QMessageBox.StandardButton.Yes:
                self.tagger.mb_login(self.on_mb_login_finished)
        else:
            dialog = PasswordDialog(authenticator, reply, parent=self)
            dialog.exec_()

    def on_mb_login_finished(self, successful, error_msg):
        if successful:
            log.debug("MusicBrainz authentication finished successfully")
        else:
            log.info("MusicBrainz authentication failed: %s", error_msg)
            QtWidgets.QMessageBox.critical(self,
                _("Authentication failed"),
                _("Login failed: %s") % error_msg)

    def show_proxy_dialog(self, proxy, authenticator):
        dialog = ProxyDialog(authenticator, proxy, parent=self)
        dialog.exec_()

    def autotag(self):
        self.tagger.autotag(self.selected_objects)

    def copy_files(self, objects):
        mimeData = QtCore.QMimeData()
        mimeData.setUrls(QtCore.QUrl.fromLocalFile(f.filename) for f in iter_files_from_objects(objects))
        self.tagger.clipboard().setMimeData(mimeData)

    def paste_files(self, target):
        mimeData = self.tagger.clipboard().mimeData()
        if mimeData.hasUrls():
            BaseTreeView.drop_urls(mimeData.urls(), target)

    def cut(self):
        self.copy_files(self.selected_objects)
        self.paste_action.setEnabled(bool(self.selected_objects))

    def paste(self):
        selected_objects = self.selected_objects
        if not selected_objects:
            target = self.tagger.unclustered_files
        else:
            target = selected_objects[0]
        self.paste_files(target)
        self.paste_action.setEnabled(False)

    def do_update_check(self):
        self.check_for_update(True)

    def auto_update_check(self):
        config = get_config()
        check_for_updates = config.setting['check_for_updates']
        update_check_days = config.setting['update_check_days']
        last_update_check = config.persist['last_update_check']
        update_level = config.setting['update_level']
        today = datetime.date.today().toordinal()
        do_auto_update_check = check_for_updates and update_check_days > 0 and today >= last_update_check + update_check_days
        log.debug("%(check_status)s startup check for program updates.  Today: %(today_date)s, Last check: %(last_check)s (Check interval: %(check_interval)s days), Update level: %(update_level)s (%(update_level_name)s)", {
            'check_status': 'Initiating' if do_auto_update_check else 'Skipping',
            'today_date': datetime.date.today(),
            'last_check': str(datetime.date.fromordinal(last_update_check)) if last_update_check > 0 else 'never',
            'check_interval': update_check_days,
            'update_level': update_level,
            'update_level_name': PROGRAM_UPDATE_LEVELS[update_level]['name'] if update_level in PROGRAM_UPDATE_LEVELS else 'unknown',
        })
        if do_auto_update_check:
            self.check_for_update(False)

    def check_for_update(self, show_always):
        config = get_config()
        self.tagger.updatecheckmanager.check_update(
            show_always=show_always,
            update_level=config.setting['update_level'],
            callback=update_last_check_date
        )

    def check_and_repair_naming_scripts(self):
        """Check the 'file_renaming_scripts' config setting to ensure that the list of scripts
        is not empty.  Check that the 'selected_file_naming_script_id' config setting points to
        a valid file naming script.
        """
        config = get_config()
        script_key = 'file_renaming_scripts'
        if not config.setting[script_key]:
            config.setting[script_key] = {
                script['id']: script.to_dict()
                for script in get_file_naming_script_presets()
            }
        naming_script_ids = list(config.setting[script_key])
        script_id_key = 'selected_file_naming_script_id'
        if config.setting[script_id_key] not in naming_script_ids:
            config.setting[script_id_key] = naming_script_ids[0]

    def check_and_repair_profiles(self):
        """Check the profiles and profile settings and repair the values if required.
        Checks that there is a settings dictionary for each profile, and that no profiles
        reference a non-existant file naming script.
        """
        script_id_key = 'selected_file_naming_script_id'
        config = get_config()
        naming_scripts = config.setting['file_renaming_scripts']
        naming_script_ids = set(naming_scripts.keys())
        profile_settings = deepcopy(config.profiles[SettingConfigSection.SETTINGS_KEY])
        for profile in config.profiles[SettingConfigSection.PROFILES_KEY]:
            p_id = profile['id']
            # Add empty settings if none found for a profile
            if p_id not in profile_settings:
                log.warning(
                    "No settings dict found for profile '%s' ('%s'). Adding empty dict.",
                    p_id,
                    profile['title'],
                )
                profile_settings[p_id] = {}
            # Remove any invalid naming script ids from profiles
            if script_id_key in profile_settings[p_id]:
                if profile_settings[p_id][script_id_key] not in naming_script_ids:
                    log.warning(
                        "Removing invalid naming script id '%s' from profile '%s' ('%s')",
                        profile_settings[p_id][script_id_key],
                        p_id,
                        profile['title'],
                    )
                    profile_settings[p_id][script_id_key] = None
        config.profiles[SettingConfigSection.SETTINGS_KEY] = profile_settings

    def make_script_selector_menu(self):
        """Update the sub-menu of available file naming scripts.
        """
        if self.script_editor_dialog is None or not isinstance(self.script_editor_dialog, ScriptEditorDialog):
            config = get_config()
            naming_scripts = config.setting['file_renaming_scripts']
            selected_script_id = config.setting['selected_file_naming_script_id']
        else:
            naming_scripts = self.script_editor_dialog.naming_scripts
            selected_script_id = self.script_editor_dialog.selected_script_id

        self.script_quick_selector_menu.clear()

        group = QtWidgets.QActionGroup(self.script_quick_selector_menu)
        group.setExclusive(True)

        def _add_menu_item(title, id):
            script_action = QtWidgets.QAction(title, self.script_quick_selector_menu)
            script_action.triggered.connect(partial(self.select_new_naming_script, id))
            script_action.setCheckable(True)
            script_action.setChecked(id == selected_script_id)
            self.script_quick_selector_menu.addAction(script_action)
            group.addAction(script_action)

        for (id, naming_script) in sorted(naming_scripts.items(), key=lambda item: item[1]['title']):
            _add_menu_item(naming_script['title'], id)

    def select_new_naming_script(self, id):
        """Update the currently selected naming script ID in the settings.

        Args:
            id (str): ID of the selected file naming script
        """
        config = get_config()
        log.debug("Setting naming script to: %s", id)
        config.setting['selected_file_naming_script_id'] = id
        self.make_script_selector_menu()
        if self.script_editor_dialog:
            self.script_editor_dialog.set_selected_script_id(id)

    def open_file_naming_script_editor(self):
        """Open the file naming script editor / manager in a new window.
        """
        self.examples = ScriptEditorExamples(tagger=self.tagger)
        self.script_editor_dialog = ScriptEditorDialog.show_instance(parent=self, examples=self.examples)
        self.script_editor_dialog.signal_save.connect(self.script_editor_save)
        self.script_editor_dialog.signal_selection_changed.connect(self.update_selector_from_script_editor)
        self.script_editor_dialog.signal_index_changed.connect(self.script_editor_index_changed)
        self.script_editor_dialog.finished.connect(self.script_editor_closed)
        # Create list of signals to disconnect when opening Options dialog.
        # Do not include `finished` because that is still used to clean up
        # locally when the editor is closed from the Options dialog.
        self.script_editor_signals = [
            self.script_editor_dialog.signal_save,
            self.script_editor_dialog.signal_selection_changed,
            self.script_editor_dialog.signal_index_changed,
        ]
        self.show_script_editor_action.setEnabled(False)

    def script_editor_save(self):
        """Process "signal_save" signal from the script editor.
        """
        self.make_script_selector_menu()

    def script_editor_closed(self):
        """Process "finished" signal from the script editor.
        """
        self.show_script_editor_action.setEnabled(True)
        self.script_editor_dialog = None
        self.make_script_selector_menu()

    def update_script_editor_example_files(self):
        """Update the list of example files for the file naming script editor.
        """
        if self.examples:
            self.examples.update_sample_example_files()
            self.update_script_editor_examples()

    def update_script_editor_examples(self):
        """Update the examples for the file naming script editor, using current settings.
        """
        if self.examples:
            config = get_config()
            override = {
                "rename_files": config.setting["rename_files"],
                "move_files": config.setting["move_files"],
            }
            self.examples.update_examples(override=override)
            if self.script_editor_dialog:
                self.script_editor_dialog.display_examples()

    def script_editor_index_changed(self):
        """Process "signal_index_changed" signal from the script editor.
        """
        self.script_editor_save()

    def update_selector_from_script_editor(self):
        """Process "signal_selection_changed" signal from the script editor.
        """
        self.script_editor_save()

    def make_profile_selector_menu(self):
        """Update the sub-menu of available option profiles.
        """
        config = get_config()
        option_profiles = config.profiles[SettingConfigSection.PROFILES_KEY]
        if not option_profiles:
            self.profile_quick_selector_menu.setDisabled(True)
            return
        self.profile_quick_selector_menu.setDisabled(False)
        self.profile_quick_selector_menu.clear()

        group = QtWidgets.QActionGroup(self.profile_quick_selector_menu)
        group.setExclusive(False)

        def _add_menu_item(title, enabled, profile_id):
            profile_action = QtWidgets.QAction(title, self.profile_quick_selector_menu)
            profile_action.triggered.connect(partial(self.update_profile_selection, profile_id))
            profile_action.setCheckable(True)
            profile_action.setChecked(enabled)
            self.profile_quick_selector_menu.addAction(profile_action)
            group.addAction(profile_action)

        for profile in option_profiles:
            _add_menu_item(profile['title'], profile['enabled'], profile['id'])

    def update_profile_selection(self, profile_id):
        """Toggle the enabled state of the selected profile.

        Args:
            profile_id (str): ID code of the profile to modify
        """
        config = get_config()
        option_profiles = config.profiles[SettingConfigSection.PROFILES_KEY]
        for profile in option_profiles:
            if profile['id'] == profile_id:
                profile['enabled'] = not profile['enabled']
                self.make_script_selector_menu()
                return

    def show_new_user_dialog(self):
        config = get_config()
        if config.setting['show_new_user_dialog']:
            msg = NewUserDialog(self)
            config.setting['show_new_user_dialog'] = msg.show()

    def check_for_plugin_update(self):
        config = get_config()
        if config.setting['check_for_plugin_updates']:
            self.tagger.pluginmanager.check_update()

    def show_plugin_update_dialog(self, plugin_names):
        if not plugin_names:
            return

        msg = PluginUpdatesDialog(self, plugin_names)
        show_options_page, perform_check = msg.show()
        config = get_config()
        config.setting['check_for_plugin_updates'] = perform_check
        if show_options_page:
            self.show_plugins_options_page()

    def show_plugins_options_page(self):
        self.show_options(page='plugins')


def update_last_check_date(is_success):
    if is_success:
        config = get_config()
        config.persist['last_update_check'] = datetime.date.today().toordinal()
    else:
        log.debug('The update check was unsuccessful. The last update date will not be changed.')



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2007 Lukáš Lalinský
# Copyright (C) 2014, 2020 Laurent Monin
# Copyright (C) 2017 Sambhav Kothari
# Copyright (C) 2019 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


# List of alias locales
ALIAS_LOCALES = {
    'aa': N_('Afar'),
    'aa_DJ': N_('Afar (Djibouti)'),
    'aa_ER': N_('Afar (Eritrea)'),
    'aa_ER_SAAHO': N_('Afar (Eritrea) (Saho)'),
    'aa_ET': N_('Afar (Ethiopia)'),
    'ab': N_('Abkhazian'),
    'ab_GE': N_('Abkhazian (Georgia)'),
    'af': N_('Afrikaans'),
    'af_NA': N_('Afrikaans (Namibia)'),
    'af_ZA': N_('Afrikaans (South Africa)'),
    'agq': N_('Aghem'),
    'agq_CM': N_('Aghem (Cameroon)'),
    'ak': N_('Akan'),
    'ak_GH': N_('Akan (Ghana)'),
    'am': N_('Amharic'),
    'am_ET': N_('Amharic (Ethiopia)'),
    'an': N_('Aragonese'),
    'an_ES': N_('Aragonese (Spain)'),
    'ann': N_('Obolo'),
    'ann_NG': N_('Obolo (Nigeria)'),
    'apc_SY': N_('Syria'),
    'ar': N_('Arabic'),
    'ar_001': N_('Arabic (world)'),
    'ar_AE': N_('Arabic (United Arab Emirates)'),
    'ar_BH': N_('Arabic (Bahrain)'),
    'ar_DJ': N_('Arabic (Djibouti)'),
    'ar_DZ': N_('Arabic (Algeria)'),
    'ar_EG': N_('Arabic (Egypt)'),
    'ar_EH': N_('Arabic (Western Sahara)'),
    'ar_ER': N_('Arabic (Eritrea)'),
    'ar_IL': N_('Arabic (Israel)'),
    'ar_IQ': N_('Arabic (Iraq)'),
    'ar_JO': N_('Arabic (Jordan)'),
    'ar_KM': N_('Arabic (Comoros)'),
    'ar_KW': N_('Arabic (Kuwait)'),
    'ar_LB': N_('Arabic (Lebanon)'),
    'ar_LY': N_('Arabic (Libya)'),
    'ar_MA': N_('Arabic (Morocco)'),
    'ar_MR': N_('Arabic (Mauritania)'),
    'ar_OM': N_('Arabic (Oman)'),
    'ar_PS': N_('Arabic (Palestinian Territories)'),
    'ar_QA': N_('Arabic (Qatar)'),
    'ar_SA': N_('Arabic (Saudi Arabia)'),
    'ar_SD': N_('Arabic (Sudan)'),
    'ar_SO': N_('Arabic (Somalia)'),
    'ar_SS': N_('Arabic (South Sudan)'),
    'ar_SY': N_('Arabic (Syria)'),
    'ar_TD': N_('Arabic (Chad)'),
    'ar_TN': N_('Arabic (Tunisia)'),
    'ar_YE': N_('Arabic (Yemen)'),
    'arn': N_('Mapuche'),
    'arn_CL': N_('Mapuche (Chile)'),
    'as': N_('Assamese'),
    'as_IN': N_('Assamese (India)'),
    'asa': N_('Asu'),
    'asa_TZ': N_('Asu (Tanzania)'),
    'ast': N_('Asturian'),
    'ast_ES': N_('Asturian (Spain)'),
    'az': N_('Azerbaijani'),
    'az_Arab': N_('Azerbaijani (Arabic)'),
    'az_Arab_IQ': N_('Azerbaijani (Arabic) (Iraq)'),
    'az_Arab_IR': N_('Azerbaijani (Arabic) (Iran)'),
    'az_Arab_TR': N_('Azerbaijani (Arabic) (Türkiye)'),
    'az_Cyrl': N_('Azerbaijani (Cyrillic)'),
    'az_Cyrl_AZ': N_('Azerbaijani (Cyrillic) (Azerbaijan)'),
    'az_Latn': N_('Azerbaijani (Latin)'),
    'az_Latn_AZ': N_('Azerbaijani (Latin) (Azerbaijan)'),
    'ba': N_('Bashkir'),
    'ba_RU': N_('Bashkir (Russia)'),
    'bal': N_('Baluchi'),
    'bal_Arab': N_('Baluchi (Arabic)'),
    'bal_Arab_PK': N_('Baluchi (Arabic) (Pakistan)'),
    'bal_Latn': N_('Baluchi (Latin)'),
    'bal_Latn_PK': N_('Baluchi (Latin) (Pakistan)'),
    'bas': N_('Basaa'),
    'bas_CM': N_('Basaa (Cameroon)'),
    'be': N_('Belarusian'),
    'be_BY': N_('Belarusian (Belarus)'),
    'be_tarask': N_('Belarusian (Taraskievica orthography)'),
    'bem': N_('Bemba'),
    'bem_ZM': N_('Bemba (Zambia)'),
    'bez': N_('Bena'),
    'bez_TZ': N_('Bena (Tanzania)'),
    'bg': N_('Bulgarian'),
    'bg_BG': N_('Bulgarian (Bulgaria)'),
    'bgc': N_('Haryanvi'),
    'bgc_IN': N_('Haryanvi (India)'),
    'bgn': N_('Western Balochi'),
    'bgn_AE': N_('Western Balochi (United Arab Emirates)'),
    'bgn_AF': N_('Western Balochi (Afghanistan)'),
    'bgn_IR': N_('Western Balochi (Iran)'),
    'bgn_OM': N_('Western Balochi (Oman)'),
    'bgn_PK': N_('Western Balochi (Pakistan)'),
    'bho': N_('Bhojpuri'),
    'bho_IN': N_('Bhojpuri (India)'),
    'blt': N_('Tai Dam'),
    'blt_VN': N_('Tai Dam (Vietnam)'),
    'bm': N_('Bambara'),
    'bm_ML': N_('Bambara (Mali)'),
    'bm_Nkoo': N_('Bambara (N’Ko)'),
    'bm_Nkoo_ML': N_('Bambara (N’Ko) (Mali)'),
    'bn': N_('Bangla'),
    'bn_BD': N_('Bangla (Bangladesh)'),
    'bn_IN': N_('Bangla (India)'),
    'bo': N_('Tibetan'),
    'bo_CN': N_('Tibetan (China)'),
    'bo_IN': N_('Tibetan (India)'),
    'br': N_('Breton'),
    'br_FR': N_('Breton (France)'),
    'brx': N_('Bodo'),
    'brx_IN': N_('Bodo (India)'),
    'bs': N_('Bosnian'),
    'bs_Cyrl': N_('Bosnian (Cyrillic)'),
    'bs_Cyrl_BA': N_('Bosnian (Cyrillic) (Bosnia & Herzegovina)'),
    'bs_Latn': N_('Bosnian (Latin)'),
    'bs_Latn_BA': N_('Bosnian (Latin) (Bosnia & Herzegovina)'),
    'bss': N_('Akoose'),
    'bss_CM': N_('Akoose (Cameroon)'),
    'byn': N_('Blin'),
    'byn_ER': N_('Blin (Eritrea)'),
    'ca': N_('Catalan'),
    'ca_AD': N_('Catalan (Andorra)'),
    'ca_ES': N_('Catalan (Spain)'),
    'ca_ES_valencia': N_('Catalan (Spain Valencian)'),
    'ca_FR': N_('Catalan (France)'),
    'ca_IT': N_('Catalan (Italy)'),
    'cad': N_('Caddo'),
    'cad_US': N_('Caddo (United States)'),
    'cch': N_('Atsam'),
    'cch_NG': N_('Atsam (Nigeria)'),
    'ccp': N_('Chakma'),
    'ccp_BD': N_('Chakma (Bangladesh)'),
    'ccp_IN': N_('Chakma (India)'),
    'ce': N_('Chechen'),
    'ce_RU': N_('Chechen (Russia)'),
    'ceb': N_('Cebuano'),
    'ceb_PH': N_('Cebuano (Philippines)'),
    'cgg': N_('Chiga'),
    'cgg_UG': N_('Chiga (Uganda)'),
    'cho': N_('Choctaw'),
    'cho_US': N_('Choctaw (United States)'),
    'chr': N_('Cherokee'),
    'chr_US': N_('Cherokee (United States)'),
    'cic': N_('Chickasaw'),
    'cic_US': N_('Chickasaw (United States)'),
    'ckb': N_('Central Kurdish'),
    'ckb_IQ': N_('Central Kurdish (Iraq)'),
    'ckb_IR': N_('Central Kurdish (Iran)'),
    'co': N_('Corsican'),
    'co_FR': N_('Corsican (France)'),
    'cs': N_('Czech'),
    'cs_CZ': N_('Czech (Czechia)'),
    'cu': N_('Church Slavic'),
    'cu_RU': N_('Church Slavic (Russia)'),
    'cv': N_('Chuvash'),
    'cv_RU': N_('Chuvash (Russia)'),
    'cy': N_('Welsh'),
    'cy_GB': N_('Welsh (United Kingdom)'),
    'da': N_('Danish'),
    'da_DK': N_('Danish (Denmark)'),
    'da_GL': N_('Danish (Greenland)'),
    'dav': N_('Taita'),
    'dav_KE': N_('Taita (Kenya)'),
    'de': N_('German'),
    'de_AT': N_('German (Austria)'),
    'de_BE': N_('German (Belgium)'),
    'de_CH': N_('German (Switzerland)'),
    'de_DE': N_('German (Germany)'),
    'de_IT': N_('German (Italy)'),
    'de_LI': N_('German (Liechtenstein)'),
    'de_LU': N_('German (Luxembourg)'),
    'dje': N_('Zarma'),
    'dje_NE': N_('Zarma (Niger)'),
    'doi': N_('Dogri'),
    'doi_IN': N_('Dogri (India)'),
    'dsb': N_('Lower Sorbian'),
    'dsb_DE': N_('Lower Sorbian (Germany)'),
    'dua': N_('Duala'),
    'dua_CM': N_('Duala (Cameroon)'),
    'dv': N_('Divehi'),
    'dv_MV': N_('Divehi (Maldives)'),
    'dyo': N_('Jola-Fonyi'),
    'dyo_SN': N_('Jola-Fonyi (Senegal)'),
    'dz': N_('Dzongkha'),
    'dz_BT': N_('Dzongkha (Bhutan)'),
    'ebu': N_('Embu'),
    'ebu_KE': N_('Embu (Kenya)'),
    'ee': N_('Ewe'),
    'ee_GH': N_('Ewe (Ghana)'),
    'ee_TG': N_('Ewe (Togo)'),
    'el': N_('Greek'),
    'el_CY': N_('Greek (Cyprus)'),
    'el_GR': N_('Greek (Greece)'),
    'el_polyton': N_('Greek (Polytonic)'),
    'en': N_('English'),
    'en_001': N_('English (world)'),
    'en_150': N_('English (Europe)'),
    'en_AE': N_('English (United Arab Emirates)'),
    'en_AG': N_('English (Antigua & Barbuda)'),
    'en_AI': N_('English (Anguilla)'),
    'en_AS': N_('English (American Samoa)'),
    'en_AT': N_('English (Austria)'),
    'en_AU': N_('English (Australia)'),
    'en_BB': N_('English (Barbados)'),
    'en_BE': N_('English (Belgium)'),
    'en_BI': N_('English (Burundi)'),
    'en_BM': N_('English (Bermuda)'),
    'en_BS': N_('English (Bahamas)'),
    'en_BW': N_('English (Botswana)'),
    'en_BZ': N_('English (Belize)'),
    'en_CA': N_('English (Canada)'),
    'en_CC': N_('English (Cocos (Keeling) Islands)'),
    'en_CH': N_('English (Switzerland)'),
    'en_CK': N_('English (Cook Islands)'),
    'en_CM': N_('English (Cameroon)'),
    'en_CX': N_('English (Christmas Island)'),
    'en_CY': N_('English (Cyprus)'),
    'en_DE': N_('English (Germany)'),
    'en_DG': N_('English (Diego Garcia)'),
    'en_DK': N_('English (Denmark)'),
    'en_DM': N_('English (Dominica)'),
    'en_Dsrt': N_('English (Deseret)'),
    'en_Dsrt_US': N_('English (Deseret) (United States)'),
    'en_ER': N_('English (Eritrea)'),
    'en_FI': N_('English (Finland)'),
    'en_FJ': N_('English (Fiji)'),
    'en_FK': N_('English (Falkland Islands)'),
    'en_FM': N_('English (Micronesia)'),
    'en_GB': N_('English (United Kingdom)'),
    'en_GD': N_('English (Grenada)'),
    'en_GG': N_('English (Guernsey)'),
    'en_GH': N_('English (Ghana)'),
    'en_GI': N_('English (Gibraltar)'),
    'en_GM': N_('English (Gambia)'),
    'en_GU': N_('English (Guam)'),
    'en_GY': N_('English (Guyana)'),
    'en_HK': N_('English (Hong Kong)'),
    'en_IE': N_('English (Ireland)'),
    'en_IL': N_('English (Israel)'),
    'en_IM': N_('English (Isle of Man)'),
    'en_IN': N_('English (India)'),
    'en_IO': N_('English (British Indian Ocean Territory)'),
    'en_JE': N_('English (Jersey)'),
    'en_JM': N_('English (Jamaica)'),
    'en_KE': N_('English (Kenya)'),
    'en_KI': N_('English (Kiribati)'),
    'en_KN': N_('English (St. Kitts & Nevis)'),
    'en_KY': N_('English (Cayman Islands)'),
    'en_LC': N_('English (St. Lucia)'),
    'en_LR': N_('English (Liberia)'),
    'en_LS': N_('English (Lesotho)'),
    'en_MG': N_('English (Madagascar)'),
    'en_MH': N_('English (Marshall Islands)'),
    'en_MO': N_('English (Macao)'),
    'en_MP': N_('English (Northern Mariana Islands)'),
    'en_MS': N_('English (Montserrat)'),
    'en_MT': N_('English (Malta)'),
    'en_MU': N_('English (Mauritius)'),
    'en_MV': N_('English (Maldives)'),
    'en_MW': N_('English (Malawi)'),
    'en_MY': N_('English (Malaysia)'),
    'en_NA': N_('English (Namibia)'),
    'en_NF': N_('English (Norfolk Island)'),
    'en_NG': N_('English (Nigeria)'),
    'en_NL': N_('English (Netherlands)'),
    'en_NR': N_('English (Nauru)'),
    'en_NU': N_('English (Niue)'),
    'en_NZ': N_('English (New Zealand)'),
    'en_PG': N_('English (Papua New Guinea)'),
    'en_PH': N_('English (Philippines)'),
    'en_PK': N_('English (Pakistan)'),
    'en_PN': N_('English (Pitcairn Islands)'),
    'en_PR': N_('English (Puerto Rico)'),
    'en_PW': N_('English (Palau)'),
    'en_RW': N_('English (Rwanda)'),
    'en_SB': N_('English (Solomon Islands)'),
    'en_SC': N_('English (Seychelles)'),
    'en_SD': N_('English (Sudan)'),
    'en_SE': N_('English (Sweden)'),
    'en_SG': N_('English (Singapore)'),
    'en_SH': N_('English (St. Helena)'),
    'en_SI': N_('English (Slovenia)'),
    'en_SL': N_('English (Sierra Leone)'),
    'en_SS': N_('English (South Sudan)'),
    'en_SX': N_('English (Sint Maarten)'),
    'en_SZ': N_('English (Eswatini)'),
    'en_Shaw': N_('English (Shavian)'),
    'en_Shaw_GB': N_('English (Shavian) (United Kingdom)'),
    'en_TC': N_('English (Turks & Caicos Islands)'),
    'en_TK': N_('English (Tokelau)'),
    'en_TO': N_('English (Tonga)'),
    'en_TT': N_('English (Trinidad & Tobago)'),
    'en_TV': N_('English (Tuvalu)'),
    'en_TZ': N_('English (Tanzania)'),
    'en_UG': N_('English (Uganda)'),
    'en_UM': N_('English (U.S. Outlying Islands)'),
    'en_US': N_('English (United States)'),
    'en_VC': N_('English (St. Vincent & Grenadines)'),
    'en_VG': N_('English (British Virgin Islands)'),
    'en_VI': N_('English (U.S. Virgin Islands)'),
    'en_VU': N_('English (Vanuatu)'),
    'en_WS': N_('English (Samoa)'),
    'en_ZA': N_('English (South Africa)'),
    'en_ZM': N_('English (Zambia)'),
    'en_ZW': N_('English (Zimbabwe)'),
    'eo': N_('Esperanto'),
    'eo_001': N_('Esperanto (world)'),
    'es': N_('Spanish'),
    'es_419': N_('Spanish (Latin America)'),
    'es_AR': N_('Spanish (Argentina)'),
    'es_BO': N_('Spanish (Bolivia)'),
    'es_BR': N_('Spanish (Brazil)'),
    'es_BZ': N_('Spanish (Belize)'),
    'es_CL': N_('Spanish (Chile)'),
    'es_CO': N_('Spanish (Colombia)'),
    'es_CR': N_('Spanish (Costa Rica)'),
    'es_CU': N_('Spanish (Cuba)'),
    'es_DO': N_('Spanish (Dominican Republic)'),
    'es_EA': N_('Spanish (Ceuta & Melilla)'),
    'es_EC': N_('Spanish (Ecuador)'),
    'es_ES': N_('Spanish (Spain)'),
    'es_GQ': N_('Spanish (Equatorial Guinea)'),
    'es_GT': N_('Spanish (Guatemala)'),
    'es_HN': N_('Spanish (Honduras)'),
    'es_IC': N_('Spanish (Canary Islands)'),
    'es_MX': N_('Spanish (Mexico)'),
    'es_NI': N_('Spanish (Nicaragua)'),
    'es_PA': N_('Spanish (Panama)'),
    'es_PE': N_('Spanish (Peru)'),
    'es_PH': N_('Spanish (Philippines)'),
    'es_PR': N_('Spanish (Puerto Rico)'),
    'es_PY': N_('Spanish (Paraguay)'),
    'es_SV': N_('Spanish (El Salvador)'),
    'es_US': N_('Spanish (United States)'),
    'es_UY': N_('Spanish (Uruguay)'),
    'es_VE': N_('Spanish (Venezuela)'),
    'et': N_('Estonian'),
    'et_EE': N_('Estonian (Estonia)'),
    'eu': N_('Basque'),
    'eu_ES': N_('Basque (Spain)'),
    'ewo': N_('Ewondo'),
    'ewo_CM': N_('Ewondo (Cameroon)'),
    'fa': N_('Persian'),
    'fa_AF': N_('Persian (Afghanistan)'),
    'fa_IR': N_('Persian (Iran)'),
    'ff': N_('Fula'),
    'ff_Adlm': N_('Fula (Adlam)'),
    'ff_Adlm_BF': N_('Fula (Adlam) (Burkina Faso)'),
    'ff_Adlm_CM': N_('Fula (Adlam) (Cameroon)'),
    'ff_Adlm_GH': N_('Fula (Adlam) (Ghana)'),
    'ff_Adlm_GM': N_('Fula (Adlam) (Gambia)'),
    'ff_Adlm_GN': N_('Fula (Adlam) (Guinea)'),
    'ff_Adlm_GW': N_('Fula (Adlam) (Guinea-Bissau)'),
    'ff_Adlm_LR': N_('Fula (Adlam) (Liberia)'),
    'ff_Adlm_MR': N_('Fula (Adlam) (Mauritania)'),
    'ff_Adlm_NE': N_('Fula (Adlam) (Niger)'),
    'ff_Adlm_NG': N_('Fula (Adlam) (Nigeria)'),
    'ff_Adlm_SL': N_('Fula (Adlam) (Sierra Leone)'),
    'ff_Adlm_SN': N_('Fula (Adlam) (Senegal)'),
    'ff_Latn': N_('Fula (Latin)'),
    'ff_Latn_BF': N_('Fula (Latin) (Burkina Faso)'),
    'ff_Latn_CM': N_('Fula (Latin) (Cameroon)'),
    'ff_Latn_GH': N_('Fula (Latin) (Ghana)'),
    'ff_Latn_GM': N_('Fula (Latin) (Gambia)'),
    'ff_Latn_GN': N_('Fula (Latin) (Guinea)'),
    'ff_Latn_GW': N_('Fula (Latin) (Guinea-Bissau)'),
    'ff_Latn_LR': N_('Fula (Latin) (Liberia)'),
    'ff_Latn_MR': N_('Fula (Latin) (Mauritania)'),
    'ff_Latn_NE': N_('Fula (Latin) (Niger)'),
    'ff_Latn_NG': N_('Fula (Latin) (Nigeria)'),
    'ff_Latn_SL': N_('Fula (Latin) (Sierra Leone)'),
    'ff_Latn_SN': N_('Fula (Latin) (Senegal)'),
    'fi': N_('Finnish'),
    'fi_FI': N_('Finnish (Finland)'),
    'fil': N_('Filipino'),
    'fil_PH': N_('Filipino (Philippines)'),
    'fo': N_('Faroese'),
    'fo_DK': N_('Faroese (Denmark)'),
    'fo_FO': N_('Faroese (Faroe Islands)'),
    'fr': N_('French'),
    'fr_BE': N_('French (Belgium)'),
    'fr_BF': N_('French (Burkina Faso)'),
    'fr_BI': N_('French (Burundi)'),
    'fr_BJ': N_('French (Benin)'),
    'fr_BL': N_('French (St. Barthélemy)'),
    'fr_CA': N_('French (Canada)'),
    'fr_CD': N_('French (Congo - Kinshasa)'),
    'fr_CF': N_('French (Central African Republic)'),
    'fr_CG': N_('French (Congo - Brazzaville)'),
    'fr_CH': N_('French (Switzerland)'),
    'fr_CI': N_('French (Côte d’Ivoire)'),
    'fr_CM': N_('French (Cameroon)'),
    'fr_DJ': N_('French (Djibouti)'),
    'fr_DZ': N_('French (Algeria)'),
    'fr_FR': N_('French (France)'),
    'fr_GA': N_('French (Gabon)'),
    'fr_GF': N_('French (French Guiana)'),
    'fr_GN': N_('French (Guinea)'),
    'fr_GP': N_('French (Guadeloupe)'),
    'fr_GQ': N_('French (Equatorial Guinea)'),
    'fr_HT': N_('French (Haiti)'),
    'fr_KM': N_('French (Comoros)'),
    'fr_LU': N_('French (Luxembourg)'),
    'fr_MA': N_('French (Morocco)'),
    'fr_MC': N_('French (Monaco)'),
    'fr_MF': N_('French (St. Martin)'),
    'fr_MG': N_('French (Madagascar)'),
    'fr_ML': N_('French (Mali)'),
    'fr_MQ': N_('French (Martinique)'),
    'fr_MR': N_('French (Mauritania)'),
    'fr_MU': N_('French (Mauritius)'),
    'fr_NC': N_('French (New Caledonia)'),
    'fr_NE': N_('French (Niger)'),
    'fr_PF': N_('French (French Polynesia)'),
    'fr_PM': N_('French (St. Pierre & Miquelon)'),
    'fr_RE': N_('French (Réunion)'),
    'fr_RW': N_('French (Rwanda)'),
    'fr_SC': N_('French (Seychelles)'),
    'fr_SN': N_('French (Senegal)'),
    'fr_SY': N_('French (Syria)'),
    'fr_TD': N_('French (Chad)'),
    'fr_TG': N_('French (Togo)'),
    'fr_TN': N_('French (Tunisia)'),
    'fr_VU': N_('French (Vanuatu)'),
    'fr_WF': N_('French (Wallis & Futuna)'),
    'fr_YT': N_('French (Mayotte)'),
    'frr': N_('Northern Frisian'),
    'frr_DE': N_('Northern Frisian (Germany)'),
    'fur': N_('Friulian'),
    'fur_IT': N_('Friulian (Italy)'),
    'fy': N_('Western Frisian'),
    'fy_NL': N_('Western Frisian (Netherlands)'),
    'ga': N_('Irish'),
    'ga_GB': N_('Irish (United Kingdom)'),
    'ga_IE': N_('Irish (Ireland)'),
    'gaa': N_('Ga'),
    'gaa_GH': N_('Ga (Ghana)'),
    'gd': N_('Scottish Gaelic'),
    'gd_GB': N_('Scottish Gaelic (United Kingdom)'),
    'gez': N_('Geez'),
    'gez_ER': N_('Geez (Eritrea)'),
    'gez_ET': N_('Geez (Ethiopia)'),
    'gl': N_('Galician'),
    'gl_ES': N_('Galician (Spain)'),
    'gn': N_('Guarani'),
    'gn_PY': N_('Guarani (Paraguay)'),
    'gsw': N_('Swiss German'),
    'gsw_CH': N_('Swiss German (Switzerland)'),
    'gsw_FR': N_('Swiss German (France)'),
    'gsw_LI': N_('Swiss German (Liechtenstein)'),
    'gu': N_('Gujarati'),
    'gu_IN': N_('Gujarati (India)'),
    'guz': N_('Gusii'),
    'guz_KE': N_('Gusii (Kenya)'),
    'gv': N_('Manx'),
    'gv_IM': N_('Manx (Isle of Man)'),
    'ha': N_('Hausa'),
    'ha_Arab': N_('Hausa (Arabic)'),
    'ha_Arab_NG': N_('Hausa (Arabic) (Nigeria)'),
    'ha_Arab_SD': N_('Hausa (Arabic) (Sudan)'),
    'ha_GH': N_('Hausa (Ghana)'),
    'ha_Latn': N_('Hausa (Latin)'),
    'ha_Latn_GH': N_('Hausa (Latin) (Ghana)'),
    'ha_Latn_NE': N_('Hausa (Latin) (Niger)'),
    'ha_Latn_NG': N_('Hausa (Latin) (Nigeria)'),
    'ha_NE': N_('Hausa (Niger)'),
    'ha_NG': N_('Hausa (Nigeria)'),
    'ha_SD': N_('Hausa (Sudan)'),
    'haw': N_('Hawaiian'),
    'haw_US': N_('Hawaiian (United States)'),
    'he': N_('Hebrew'),
    'he_IL': N_('Hebrew (Israel)'),
    'hi': N_('Hindi'),
    'hi_IN': N_('Hindi (India)'),
    'hi_Latn': N_('Hindi (Latin)'),
    'hi_Latn_IN': N_('Hindi (Latin) (India)'),
    'hnj': N_('Hmong Njua'),
    'hnj_Hmnp': N_('Hmong Njua (Nyiakeng Puachue Hmong)'),
    'hnj_Hmnp_US': N_('Hmong Njua (Nyiakeng Puachue Hmong) (United States)'),
    'hr': N_('Croatian'),
    'hr_BA': N_('Croatian (Bosnia & Herzegovina)'),
    'hr_HR': N_('Croatian (Croatia)'),
    'hsb': N_('Upper Sorbian'),
    'hsb_DE': N_('Upper Sorbian (Germany)'),
    'hu': N_('Hungarian'),
    'hu_HU': N_('Hungarian (Hungary)'),
    'hy': N_('Armenian'),
    'hy_AM': N_('Armenian (Armenia)'),
    'hy_AM_REVISED': N_('Armenian (Armenia) (Revised Orthography)'),
    'ia': N_('Interlingua'),
    'ia_001': N_('Interlingua (world)'),
    'id': N_('Indonesian'),
    'id_ID': N_('Indonesian (Indonesia)'),
    'ig': N_('Igbo'),
    'ig_NG': N_('Igbo (Nigeria)'),
    'ii': N_('Sichuan Yi'),
    'ii_CN': N_('Sichuan Yi (China)'),
    'io': N_('Ido'),
    'io_001': N_('Ido (world)'),
    'is': N_('Icelandic'),
    'is_IS': N_('Icelandic (Iceland)'),
    'it': N_('Italian'),
    'it_CH': N_('Italian (Switzerland)'),
    'it_IT': N_('Italian (Italy)'),
    'it_SM': N_('Italian (San Marino)'),
    'it_VA': N_('Italian (Vatican City)'),
    'iu': N_('Inuktitut'),
    'iu_CA': N_('Inuktitut (Canada)'),
    'iu_Latn': N_('Inuktitut (Latin)'),
    'iu_Latn_CA': N_('Inuktitut (Latin) (Canada)'),
    'ja': N_('Japanese'),
    'ja_JP': N_('Japanese (Japan)'),
    'jbo': N_('Lojban'),
    'jbo_001': N_('Lojban (world)'),
    'jgo': N_('Ngomba'),
    'jgo_CM': N_('Ngomba (Cameroon)'),
    'jmc': N_('Machame'),
    'jmc_TZ': N_('Machame (Tanzania)'),
    'jv': N_('Javanese'),
    'jv_ID': N_('Javanese (Indonesia)'),
    'ka': N_('Georgian'),
    'ka_GE': N_('Georgian (Georgia)'),
    'kab': N_('Kabyle'),
    'kab_DZ': N_('Kabyle (Algeria)'),
    'kaj': N_('Jju'),
    'kaj_NG': N_('Jju (Nigeria)'),
    'kam': N_('Kamba'),
    'kam_KE': N_('Kamba (Kenya)'),
    'kcg': N_('Tyap'),
    'kcg_NG': N_('Tyap (Nigeria)'),
    'kde': N_('Makonde'),
    'kde_TZ': N_('Makonde (Tanzania)'),
    'kea': N_('Kabuverdianu'),
    'kea_CV': N_('Kabuverdianu (Cape Verde)'),
    'ken': N_('Kenyang'),
    'ken_CM': N_('Kenyang (Cameroon)'),
    'kfo': N_('Koro'),
    'kfo_CI': N_('Koro (Côte d’Ivoire)'),
    'kgp': N_('Kaingang'),
    'kgp_BR': N_('Kaingang (Brazil)'),
    'khq': N_('Koyra Chiini'),
    'khq_ML': N_('Koyra Chiini (Mali)'),
    'ki': N_('Kikuyu'),
    'ki_KE': N_('Kikuyu (Kenya)'),
    'kk': N_('Kazakh'),
    'kk_Cyrl': N_('Kazakh (Cyrillic)'),
    'kk_Cyrl_KZ': N_('Kazakh (Cyrillic) (Kazakhstan)'),
    'kk_KZ': N_('Kazakh (Kazakhstan)'),
    'kkj': N_('Kako'),
    'kkj_CM': N_('Kako (Cameroon)'),
    'kl': N_('Kalaallisut'),
    'kl_GL': N_('Kalaallisut (Greenland)'),
    'kln': N_('Kalenjin'),
    'kln_KE': N_('Kalenjin (Kenya)'),
    'km': N_('Khmer'),
    'km_KH': N_('Khmer (Cambodia)'),
    'kn': N_('Kannada'),
    'kn_IN': N_('Kannada (India)'),
    'ko': N_('Korean'),
    'ko_KP': N_('Korean (North Korea)'),
    'ko_KR': N_('Korean (South Korea)'),
    'kok': N_('Konkani'),
    'kok_IN': N_('Konkani (India)'),
    'kpe': N_('Kpelle'),
    'kpe_GN': N_('Kpelle (Guinea)'),
    'kpe_LR': N_('Kpelle (Liberia)'),
    'ks': N_('Kashmiri'),
    'ks_Arab': N_('Kashmiri (Arabic)'),
    'ks_Arab_IN': N_('Kashmiri (Arabic) (India)'),
    'ks_Deva': N_('Kashmiri (Devanagari)'),
    'ks_Deva_IN': N_('Kashmiri (Devanagari) (India)'),
    'ksb': N_('Shambala'),
    'ksb_TZ': N_('Shambala (Tanzania)'),
    'ksf': N_('Bafia'),
    'ksf_CM': N_('Bafia (Cameroon)'),
    'ksh': N_('Colognian'),
    'ksh_DE': N_('Colognian (Germany)'),
    'ku': N_('Kurdish'),
    'ku_TR': N_('Kurdish (Türkiye)'),
    'kw': N_('Cornish'),
    'kw_GB': N_('Cornish (United Kingdom)'),
    'ky': N_('Kyrgyz'),
    'ky_KG': N_('Kyrgyz (Kyrgyzstan)'),
    'la': N_('Latin'),
    'la_VA': N_('Latin (Vatican City)'),
    'lag': N_('Langi'),
    'lag_TZ': N_('Langi (Tanzania)'),
    'lb': N_('Luxembourgish'),
    'lb_LU': N_('Luxembourgish (Luxembourg)'),
    'lg': N_('Ganda'),
    'lg_UG': N_('Ganda (Uganda)'),
    'lij': N_('Ligurian'),
    'lij_IT': N_('Ligurian (Italy)'),
    'lkt': N_('Lakota'),
    'lkt_US': N_('Lakota (United States)'),
    'lmo': N_('Lombard'),
    'lmo_IT': N_('Lombard (Italy)'),
    'ln': N_('Lingala'),
    'ln_AO': N_('Lingala (Angola)'),
    'ln_CD': N_('Lingala (Congo - Kinshasa)'),
    'ln_CF': N_('Lingala (Central African Republic)'),
    'ln_CG': N_('Lingala (Congo - Brazzaville)'),
    'lo': N_('Lao'),
    'lo_LA': N_('Lao (Laos)'),
    'lrc': N_('Northern Luri'),
    'lrc_IQ': N_('Northern Luri (Iraq)'),
    'lrc_IR': N_('Northern Luri (Iran)'),
    'lt': N_('Lithuanian'),
    'lt_LT': N_('Lithuanian (Lithuania)'),
    'lu': N_('Luba-Katanga'),
    'lu_CD': N_('Luba-Katanga (Congo - Kinshasa)'),
    'luo': N_('Luo'),
    'luo_KE': N_('Luo (Kenya)'),
    'luy': N_('Luyia'),
    'luy_KE': N_('Luyia (Kenya)'),
    'lv': N_('Latvian'),
    'lv_LV': N_('Latvian (Latvia)'),
    'mai': N_('Maithili'),
    'mai_IN': N_('Maithili (India)'),
    'mas': N_('Masai'),
    'mas_KE': N_('Masai (Kenya)'),
    'mas_TZ': N_('Masai (Tanzania)'),
    'mdf': N_('Moksha'),
    'mdf_RU': N_('Moksha (Russia)'),
    'mer': N_('Meru'),
    'mer_KE': N_('Meru (Kenya)'),
    'mfe': N_('Morisyen'),
    'mfe_MU': N_('Morisyen (Mauritius)'),
    'mg': N_('Malagasy'),
    'mg_MG': N_('Malagasy (Madagascar)'),
    'mgh': N_('Makhuwa-Meetto'),
    'mgh_MZ': N_('Makhuwa-Meetto (Mozambique)'),
    'mgo': N_('Metaʼ'),
    'mgo_CM': N_('Metaʼ (Cameroon)'),
    'mi': N_('Māori'),
    'mi_NZ': N_('Māori (New Zealand)'),
    'mk': N_('Macedonian'),
    'mk_MK': N_('Macedonian (North Macedonia)'),
    'ml': N_('Malayalam'),
    'ml_IN': N_('Malayalam (India)'),
    'mn': N_('Mongolian'),
    'mn_Cyrl': N_('Mongolian (Cyrillic)'),
    'mn_Cyrl_MN': N_('Mongolian (Cyrillic) (Mongolia)'),
    'mn_MN': N_('Mongolian (Mongolia)'),
    'mn_Mong': N_('Mongolian (Mongolian)'),
    'mn_Mong_CN': N_('Mongolian (Mongolian) (China)'),
    'mn_Mong_MN': N_('Mongolian (Mongolian) (Mongolia)'),
    'mni': N_('Manipuri'),
    'mni_Beng': N_('Manipuri (Bangla)'),
    'mni_Beng_IN': N_('Manipuri (Bangla) (India)'),
    'mni_Mtei': N_('Manipuri (Meitei Mayek)'),
    'mni_Mtei_IN': N_('Manipuri (Meitei Mayek) (India)'),
    'moh': N_('Mohawk'),
    'moh_CA': N_('Mohawk (Canada)'),
    'mr': N_('Marathi'),
    'mr_IN': N_('Marathi (India)'),
    'ms': N_('Malay'),
    'ms_Arab': N_('Malay (Arabic)'),
    'ms_Arab_BN': N_('Malay (Arabic) (Brunei)'),
    'ms_Arab_MY': N_('Malay (Arabic) (Malaysia)'),
    'ms_BN': N_('Malay (Brunei)'),
    'ms_ID': N_('Malay (Indonesia)'),
    'ms_MY': N_('Malay (Malaysia)'),
    'ms_SG': N_('Malay (Singapore)'),
    'mt': N_('Maltese'),
    'mt_MT': N_('Maltese (Malta)'),
    'mua': N_('Mundang'),
    'mua_CM': N_('Mundang (Cameroon)'),
    'mus': N_('Muscogee'),
    'mus_US': N_('Muscogee (United States)'),
    'my': N_('Burmese'),
    'my_MM': N_('Burmese (Myanmar (Burma))'),
    'myv': N_('Erzya'),
    'myv_RU': N_('Erzya (Russia)'),
    'mzn': N_('Mazanderani'),
    'mzn_IR': N_('Mazanderani (Iran)'),
    'naq': N_('Nama'),
    'naq_NA': N_('Nama (Namibia)'),
    'nb': N_('Norwegian Bokmål'),
    'nb_NO': N_('Norwegian Bokmål (Norway)'),
    'nb_SJ': N_('Norwegian Bokmål (Svalbard & Jan Mayen)'),
    'nd': N_('North Ndebele'),
    'nd_ZW': N_('North Ndebele (Zimbabwe)'),
    'nds': N_('Low German'),
    'nds_DE': N_('Low German (Germany)'),
    'nds_NL': N_('Low German (Netherlands)'),
    'ne': N_('Nepali'),
    'ne_IN': N_('Nepali (India)'),
    'ne_NP': N_('Nepali (Nepal)'),
    'nl': N_('Dutch'),
    'nl_AW': N_('Dutch (Aruba)'),
    'nl_BE': N_('Dutch (Belgium)'),
    'nl_BQ': N_('Dutch (Caribbean Netherlands)'),
    'nl_CW': N_('Dutch (Curaçao)'),
    'nl_NL': N_('Dutch (Netherlands)'),
    'nl_SR': N_('Dutch (Suriname)'),
    'nl_SX': N_('Dutch (Sint Maarten)'),
    'nmg': N_('Kwasio'),
    'nmg_CM': N_('Kwasio (Cameroon)'),
    'nn': N_('Norwegian Nynorsk'),
    'nn_NO': N_('Norwegian Nynorsk (Norway)'),
    'nnh': N_('Ngiemboon'),
    'nnh_CM': N_('Ngiemboon (Cameroon)'),
    'no': N_('Norwegian'),
    'nqo': N_('N’Ko'),
    'nqo_GN': N_('N’Ko (Guinea)'),
    'nr': N_('South Ndebele'),
    'nr_ZA': N_('South Ndebele (South Africa)'),
    'nso': N_('Northern Sotho'),
    'nso_ZA': N_('Northern Sotho (South Africa)'),
    'nus': N_('Nuer'),
    'nus_SS': N_('Nuer (South Sudan)'),
    'nv': N_('Navajo'),
    'nv_US': N_('Navajo (United States)'),
    'ny': N_('Nyanja'),
    'ny_MW': N_('Nyanja (Malawi)'),
    'nyn': N_('Nyankole'),
    'nyn_UG': N_('Nyankole (Uganda)'),
    'oc': N_('Occitan'),
    'oc_ES': N_('Occitan (Spain)'),
    'oc_FR': N_('Occitan (France)'),
    'om': N_('Oromo'),
    'om_ET': N_('Oromo (Ethiopia)'),
    'om_KE': N_('Oromo (Kenya)'),
    'or': N_('Odia'),
    'or_IN': N_('Odia (India)'),
    'os': N_('Ossetic'),
    'os_GE': N_('Ossetic (Georgia)'),
    'os_RU': N_('Ossetic (Russia)'),
    'osa': N_('Osage'),
    'osa_US': N_('Osage (United States)'),
    'pa': N_('Punjabi'),
    'pa_Arab': N_('Punjabi (Arabic)'),
    'pa_Arab_PK': N_('Punjabi (Arabic) (Pakistan)'),
    'pa_Guru': N_('Punjabi (Gurmukhi)'),
    'pa_Guru_IN': N_('Punjabi (Gurmukhi) (India)'),
    'pa_IN': N_('Punjabi (India)'),
    'pa_PK': N_('Punjabi (Pakistan)'),
    'pap': N_('Papiamento'),
    'pap_AW': N_('Papiamento (Aruba)'),
    'pap_CW': N_('Papiamento (Curaçao)'),
    'pcm': N_('Nigerian Pidgin'),
    'pcm_NG': N_('Nigerian Pidgin (Nigeria)'),
    'pis': N_('Pijin'),
    'pis_SB': N_('Pijin (Solomon Islands)'),
    'pl': N_('Polish'),
    'pl_PL': N_('Polish (Poland)'),
    'prg': N_('Prussian'),
    'prg_001': N_('Prussian (world)'),
    'ps': N_('Pashto'),
    'ps_AF': N_('Pashto (Afghanistan)'),
    'ps_PK': N_('Pashto (Pakistan)'),
    'pt': N_('Portuguese'),
    'pt_AO': N_('Portuguese (Angola)'),
    'pt_BR': N_('Portuguese (Brazil)'),
    'pt_CH': N_('Portuguese (Switzerland)'),
    'pt_CV': N_('Portuguese (Cape Verde)'),
    'pt_GQ': N_('Portuguese (Equatorial Guinea)'),
    'pt_GW': N_('Portuguese (Guinea-Bissau)'),
    'pt_LU': N_('Portuguese (Luxembourg)'),
    'pt_MO': N_('Portuguese (Macao)'),
    'pt_MZ': N_('Portuguese (Mozambique)'),
    'pt_PT': N_('Portuguese (Portugal)'),
    'pt_ST': N_('Portuguese (São Tomé & Príncipe)'),
    'pt_TL': N_('Portuguese (Timor-Leste)'),
    'qu': N_('Quechua'),
    'qu_BO': N_('Quechua (Bolivia)'),
    'qu_EC': N_('Quechua (Ecuador)'),
    'qu_PE': N_('Quechua (Peru)'),
    'quc': N_('Kʼicheʼ'),
    'quc_GT': N_('Kʼicheʼ (Guatemala)'),
    'raj': N_('Rajasthani'),
    'raj_IN': N_('Rajasthani (India)'),
    'rhg': N_('Rohingya'),
    'rhg_Rohg': N_('Rohingya (Hanifi)'),
    'rhg_Rohg_BD': N_('Rohingya (Hanifi) (Bangladesh)'),
    'rhg_Rohg_MM': N_('Rohingya (Hanifi) (Myanmar (Burma))'),
    'rif': N_('Riffian'),
    'rif_MA': N_('Riffian (Morocco)'),
    'rm': N_('Romansh'),
    'rm_CH': N_('Romansh (Switzerland)'),
    'rn': N_('Rundi'),
    'rn_BI': N_('Rundi (Burundi)'),
    'ro': N_('Romanian'),
    'ro_MD': N_('Romanian (Moldova)'),
    'ro_RO': N_('Romanian (Romania)'),
    'rof': N_('Rombo'),
    'rof_TZ': N_('Rombo (Tanzania)'),
    'ru': N_('Russian'),
    'ru_BY': N_('Russian (Belarus)'),
    'ru_KG': N_('Russian (Kyrgyzstan)'),
    'ru_KZ': N_('Russian (Kazakhstan)'),
    'ru_MD': N_('Russian (Moldova)'),
    'ru_RU': N_('Russian (Russia)'),
    'ru_UA': N_('Russian (Ukraine)'),
    'rw': N_('Kinyarwanda'),
    'rw_RW': N_('Kinyarwanda (Rwanda)'),
    'rwk': N_('Rwa'),
    'rwk_TZ': N_('Rwa (Tanzania)'),
    'sa': N_('Sanskrit'),
    'sa_IN': N_('Sanskrit (India)'),
    'sah': N_('Yakut'),
    'sah_RU': N_('Yakut (Russia)'),
    'saq': N_('Samburu'),
    'saq_KE': N_('Samburu (Kenya)'),
    'sat': N_('Santali'),
    'sat_Deva': N_('Santali (Devanagari)'),
    'sat_Deva_IN': N_('Santali (Devanagari) (India)'),
    'sat_Olck': N_('Santali (Ol Chiki)'),
    'sat_Olck_IN': N_('Santali (Ol Chiki) (India)'),
    'sbp': N_('Sangu'),
    'sbp_TZ': N_('Sangu (Tanzania)'),
    'sc': N_('Sardinian'),
    'sc_IT': N_('Sardinian (Italy)'),
    'scn': N_('Sicilian'),
    'scn_IT': N_('Sicilian (Italy)'),
    'sd': N_('Sindhi'),
    'sd_Arab': N_('Sindhi (Arabic)'),
    'sd_Arab_PK': N_('Sindhi (Arabic) (Pakistan)'),
    'sd_Deva': N_('Sindhi (Devanagari)'),
    'sd_Deva_IN': N_('Sindhi (Devanagari) (India)'),
    'sdh': N_('Southern Kurdish'),
    'sdh_IQ': N_('Southern Kurdish (Iraq)'),
    'sdh_IR': N_('Southern Kurdish (Iran)'),
    'se': N_('Northern Sami'),
    'se_FI': N_('Northern Sami (Finland)'),
    'se_NO': N_('Northern Sami (Norway)'),
    'se_SE': N_('Northern Sami (Sweden)'),
    'seh': N_('Sena'),
    'seh_MZ': N_('Sena (Mozambique)'),
    'ses': N_('Koyraboro Senni'),
    'ses_ML': N_('Koyraboro Senni (Mali)'),
    'sg': N_('Sango'),
    'sg_CF': N_('Sango (Central African Republic)'),
    'shi': N_('Tachelhit'),
    'shi_Latn': N_('Tachelhit (Latin)'),
    'shi_Latn_MA': N_('Tachelhit (Latin) (Morocco)'),
    'shi_Tfng': N_('Tachelhit (Tifinagh)'),
    'shi_Tfng_MA': N_('Tachelhit (Tifinagh) (Morocco)'),
    'shn': N_('Shan'),
    'shn_MM': N_('Shan (Myanmar (Burma))'),
    'shn_TH': N_('Shan (Thailand)'),
    'si': N_('Sinhala'),
    'si_LK': N_('Sinhala (Sri Lanka)'),
    'sid': N_('Sidamo'),
    'sid_ET': N_('Sidamo (Ethiopia)'),
    'sk': N_('Slovak'),
    'sk_SK': N_('Slovak (Slovakia)'),
    'sl': N_('Slovenian'),
    'sl_SI': N_('Slovenian (Slovenia)'),
    'sma': N_('Southern Sami'),
    'sma_NO': N_('Southern Sami (Norway)'),
    'sma_SE': N_('Southern Sami (Sweden)'),
    'smj': N_('Lule Sami'),
    'smj_NO': N_('Lule Sami (Norway)'),
    'smj_SE': N_('Lule Sami (Sweden)'),
    'smn': N_('Inari Sami'),
    'smn_FI': N_('Inari Sami (Finland)'),
    'sms': N_('Skolt Sami'),
    'sms_FI': N_('Skolt Sami (Finland)'),
    'sn': N_('Shona'),
    'sn_ZW': N_('Shona (Zimbabwe)'),
    'so': N_('Somali'),
    'so_DJ': N_('Somali (Djibouti)'),
    'so_ET': N_('Somali (Ethiopia)'),
    'so_KE': N_('Somali (Kenya)'),
    'so_SO': N_('Somali (Somalia)'),
    'sq': N_('Albanian'),
    'sq_AL': N_('Albanian (Albania)'),
    'sq_MK': N_('Albanian (North Macedonia)'),
    'sq_XK': N_('Albanian (Kosovo)'),
    'sr': N_('Serbian'),
    'sr_Cyrl': N_('Serbian (Cyrillic)'),
    'sr_Cyrl_BA': N_('Serbian (Cyrillic) (Bosnia & Herzegovina)'),
    'sr_Cyrl_ME': N_('Serbian (Cyrillic) (Montenegro)'),
    'sr_Cyrl_RS': N_('Serbian (Cyrillic) (Serbia)'),
    'sr_Cyrl_XK': N_('Serbian (Cyrillic) (Kosovo)'),
    'sr_Latn': N_('Serbian (Latin)'),
    'sr_Latn_BA': N_('Serbian (Latin) (Bosnia & Herzegovina)'),
    'sr_Latn_ME': N_('Serbian (Latin) (Montenegro)'),
    'sr_Latn_RS': N_('Serbian (Latin) (Serbia)'),
    'sr_Latn_XK': N_('Serbian (Latin) (Kosovo)'),
    'ss': N_('Swati'),
    'ss_SZ': N_('Swati (Eswatini)'),
    'ss_ZA': N_('Swati (South Africa)'),
    'ssy': N_('Saho'),
    'ssy_ER': N_('Saho (Eritrea)'),
    'st': N_('Southern Sotho'),
    'st_LS': N_('Southern Sotho (Lesotho)'),
    'st_ZA': N_('Southern Sotho (South Africa)'),
    'su': N_('Sundanese'),
    'su_Latn': N_('Sundanese (Latin)'),
    'su_Latn_ID': N_('Sundanese (Latin) (Indonesia)'),
    'sv': N_('Swedish'),
    'sv_AX': N_('Swedish (Åland Islands)'),
    'sv_FI': N_('Swedish (Finland)'),
    'sv_SE': N_('Swedish (Sweden)'),
    'sw': N_('Swahili'),
    'sw_CD': N_('Swahili (Congo - Kinshasa)'),
    'sw_KE': N_('Swahili (Kenya)'),
    'sw_TZ': N_('Swahili (Tanzania)'),
    'sw_UG': N_('Swahili (Uganda)'),
    'syr': N_('Syriac'),
    'syr_IQ': N_('Syriac (Iraq)'),
    'syr_SY': N_('Syriac (Syria)'),
    'szl': N_('Silesian'),
    'szl_PL': N_('Silesian (Poland)'),
    'ta': N_('Tamil'),
    'ta_IN': N_('Tamil (India)'),
    'ta_LK': N_('Tamil (Sri Lanka)'),
    'ta_MY': N_('Tamil (Malaysia)'),
    'ta_SG': N_('Tamil (Singapore)'),
    'te': N_('Telugu'),
    'te_IN': N_('Telugu (India)'),
    'teo': N_('Teso'),
    'teo_KE': N_('Teso (Kenya)'),
    'teo_UG': N_('Teso (Uganda)'),
    'tg': N_('Tajik'),
    'tg_Cyrl': N_('Tajik (Cyrillic)'),
    'tg_Cyrl_TJ': N_('Tajik (Cyrillic) (Tajikistan)'),
    'tg_TJ': N_('Tajik (Tajikistan)'),
    'th': N_('Thai'),
    'th_TH': N_('Thai (Thailand)'),
    'ti': N_('Tigrinya'),
    'ti_ER': N_('Tigrinya (Eritrea)'),
    'ti_ET': N_('Tigrinya (Ethiopia)'),
    'tig': N_('Tigre'),
    'tig_ER': N_('Tigre (Eritrea)'),
    'tk': N_('Turkmen'),
    'tk_TM': N_('Turkmen (Turkmenistan)'),
    'tn': N_('Tswana'),
    'tn_BW': N_('Tswana (Botswana)'),
    'tn_ZA': N_('Tswana (South Africa)'),
    'to': N_('Tongan'),
    'to_TO': N_('Tongan (Tonga)'),
    'tok': N_('Toki Pona'),
    'tok_001': N_('Toki Pona (world)'),
    'tpi': N_('Tok Pisin'),
    'tpi_PG': N_('Tok Pisin (Papua New Guinea)'),
    'tr': N_('Turkish'),
    'tr_CY': N_('Turkish (Cyprus)'),
    'tr_TR': N_('Turkish (Türkiye)'),
    'trv': N_('Taroko'),
    'trv_TW': N_('Taroko (Taiwan)'),
    'trw': N_('Torwali'),
    'trw_PK': N_('Torwali (Pakistan)'),
    'ts': N_('Tsonga'),
    'ts_ZA': N_('Tsonga (South Africa)'),
    'tt': N_('Tatar'),
    'tt_RU': N_('Tatar (Russia)'),
    'twq': N_('Tasawaq'),
    'twq_NE': N_('Tasawaq (Niger)'),
    'tzm': N_('Central Atlas Tamazight'),
    'tzm_MA': N_('Central Atlas Tamazight (Morocco)'),
    'ug': N_('Uyghur'),
    'ug_CN': N_('Uyghur (China)'),
    'uk': N_('Ukrainian'),
    'uk_UA': N_('Ukrainian (Ukraine)'),
    'und': N_('Unknown language'),
    'ur': N_('Urdu'),
    'ur_IN': N_('Urdu (India)'),
    'ur_PK': N_('Urdu (Pakistan)'),
    'uz': N_('Uzbek'),
    'uz_Arab': N_('Uzbek (Arabic)'),
    'uz_Arab_AF': N_('Uzbek (Arabic) (Afghanistan)'),
    'uz_Cyrl': N_('Uzbek (Cyrillic)'),
    'uz_Cyrl_UZ': N_('Uzbek (Cyrillic) (Uzbekistan)'),
    'uz_Latn': N_('Uzbek (Latin)'),
    'uz_Latn_UZ': N_('Uzbek (Latin) (Uzbekistan)'),
    'vai': N_('Vai'),
    'vai_Latn': N_('Vai (Latin)'),
    'vai_Latn_LR': N_('Vai (Latin) (Liberia)'),
    'vai_Vaii': N_('Vai (Vai)'),
    'vai_Vaii_LR': N_('Vai (Vai) (Liberia)'),
    've': N_('Venda'),
    've_ZA': N_('Venda (South Africa)'),
    'vec': N_('Venetian'),
    'vec_IT': N_('Venetian (Italy)'),
    'vi': N_('Vietnamese'),
    'vi_VN': N_('Vietnamese (Vietnam)'),
    'vo': N_('Volapük'),
    'vo_001': N_('Volapük (world)'),
    'vun': N_('Vunjo'),
    'vun_TZ': N_('Vunjo (Tanzania)'),
    'wa': N_('Walloon'),
    'wa_BE': N_('Walloon (Belgium)'),
    'wae': N_('Walser'),
    'wae_CH': N_('Walser (Switzerland)'),
    'wal': N_('Wolaytta'),
    'wal_ET': N_('Wolaytta (Ethiopia)'),
    'wbp': N_('Warlpiri'),
    'wbp_AU': N_('Warlpiri (Australia)'),
    'wo': N_('Wolof'),
    'wo_SN': N_('Wolof (Senegal)'),
    'xh': N_('Xhosa'),
    'xh_ZA': N_('Xhosa (South Africa)'),
    'xog': N_('Soga'),
    'xog_UG': N_('Soga (Uganda)'),
    'yav': N_('Yangben'),
    'yav_CM': N_('Yangben (Cameroon)'),
    'yi': N_('Yiddish'),
    'yi_001': N_('Yiddish (world)'),
    'yo': N_('Yoruba'),
    'yo_BJ': N_('Yoruba (Benin)'),
    'yo_NG': N_('Yoruba (Nigeria)'),
    'yrl': N_('Nheengatu'),
    'yrl_BR': N_('Nheengatu (Brazil)'),
    'yrl_CO': N_('Nheengatu (Colombia)'),
    'yrl_VE': N_('Nheengatu (Venezuela)'),
    'yue': N_('Cantonese'),
    'yue_Hans': N_('Cantonese (Simplified)'),
    'yue_Hans_CN': N_('Cantonese (Simplified) (China)'),
    'yue_Hant': N_('Cantonese (Traditional)'),
    'yue_Hant_HK': N_('Cantonese (Traditional) (Hong Kong)'),
    'zgh': N_('Standard Moroccan Tamazight'),
    'zgh_MA': N_('Standard Moroccan Tamazight (Morocco)'),
    'zh': N_('Chinese'),
    'zh_Hans': N_('Chinese (Simplified)'),
    'zh_Hans_CN': N_('Chinese (Simplified) (China)'),
    'zh_Hans_HK': N_('Chinese (Simplified) (Hong Kong)'),
    'zh_Hans_MO': N_('Chinese (Simplified) (Macao)'),
    'zh_Hans_SG': N_('Chinese (Simplified) (Singapore)'),
    'zh_Hant': N_('Chinese (Traditional)'),
    'zh_Hant_HK': N_('Chinese (Traditional) (Hong Kong)'),
    'zh_Hant_MO': N_('Chinese (Traditional) (Macao)'),
    'zh_Hant_TW': N_('Chinese (Traditional) (Taiwan)'),
    'zu': N_('Zulu'),
    'zu_ZA': N_('Zulu (South Africa)'),
}



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2021-2022 Laurent Monin
# Copyright (C) 2021-2022 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


import os
import os.path

from PyQt5.QtCore import (
    QCoreApplication,
    QStandardPaths,
)

from picard import (
    PICARD_APP_NAME,
    PICARD_ORG_NAME,
)


# Ensure the application is properly configured for the paths to work
QCoreApplication.setApplicationName(PICARD_APP_NAME)
QCoreApplication.setOrganizationName(PICARD_ORG_NAME)


def config_folder():
    return os.path.normpath(os.environ.get('PICARD_CONFIG_DIR', QStandardPaths.writableLocation(QStandardPaths.StandardLocation.AppConfigLocation)))


def cache_folder():
    return os.path.normpath(os.environ.get('PICARD_CACHE_DIR', QStandardPaths.writableLocation(QStandardPaths.StandardLocation.CacheLocation)))


def plugin_folder():
    # FIXME: This really should be in QStandardPaths.StandardLocation.AppDataLocation instead,
    # but this is a breaking change that requires data migration
    return os.path.normpath(os.environ.get('PICARD_PLUGIN_DIR', os.path.join(config_folder(), 'plugins')))



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2007, 2014, 2016 Lukáš Lalinský
# Copyright (C) 2014, 2019-2022 Philipp Wolfer
# Copyright (C) 2014-2016, 2018-2021, 2023 Laurent Monin
# Copyright (C) 2015 Ohm Patel
# Copyright (C) 2016 Rahul Raturi
# Copyright (C) 2016 Wieland Hoffmann
# Copyright (C) 2016-2017 Frederik “Freso” S. Olesen
# Copyright (C) 2017 Antonio Larrosa
# Copyright (C) 2017 Sambhav Kothari
# Copyright (C) 2018 Vishal Choudhary
# Copyright (C) 2018, 2021, 2023 Bob Swift
# Copyright (C) 2020 RomFouq
# Copyright (C) 2021 Gabriel Ferreira
# Copyright (C) 2021 Vladislav Karbovskii
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


import builtins
from collections import OrderedDict

from picard import PICARD_VERSION
from picard.const import appdirs
from picard.const.attributes import MB_ATTRIBUTES


# Install gettext "noop" function in case const.py gets imported directly.
builtins.__dict__['N_'] = lambda a: a


# Config directory
USER_DIR = appdirs.config_folder()
USER_PLUGIN_DIR = appdirs.plugin_folder()

# Network Cache default settings
CACHE_SIZE_IN_BYTES = 100*1000*1000

# AcoustID client API key
ACOUSTID_KEY = 'v8pQ6oyB'
ACOUSTID_URL = 'https://api.acoustid.org/v2'
FPCALC_NAMES = ['fpcalc', 'pyfpcalc']
DEFAULT_FPCALC_THREADS = 2

# MB OAuth client credentials
MUSICBRAINZ_OAUTH_CLIENT_ID = 'ACa9wsDX19cLp-AeEP-vVw'
MUSICBRAINZ_OAUTH_CLIENT_SECRET = 'xIsvXbIuntaLuRRhzuazOA'

# Cover art archive URL
CAA_URL = 'https://coverartarchive.org'

# Prepare documentation URLs
if PICARD_VERSION.identifier == 'final':
    DOCS_VERSION = "v{}.{}/".format(PICARD_VERSION.major, PICARD_VERSION.minor)
else:
    DOCS_VERSION = ""  # points to latest version
DOCS_LANGUAGE = 'en'
DOCS_SERVER_URL = "https://picard-docs.musicbrainz.org/"
DOCS_BASE_URL = DOCS_SERVER_URL + DOCS_VERSION + DOCS_LANGUAGE

# URLs
PICARD_URLS = {
    'home':                    "https://picard.musicbrainz.org/",
    'license':                 "https://www.gnu.org/licenses/gpl-2.0.html",
    'documentation_server':    DOCS_SERVER_URL,     # Shows latest version and tries to match the user's language if available.
    'documentation':           DOCS_BASE_URL + "/",
    'troubleshooting':         DOCS_BASE_URL + "/troubleshooting/troubleshooting.html",
    'doc_options':             DOCS_BASE_URL + "/config/configuration.html",
    'doc_scripting':           DOCS_BASE_URL + "/extending/scripting.html",
    'doc_tags_from_filenames': DOCS_BASE_URL + "/usage/tags_from_file_names.html",
    'doc_naming_script_edit':  DOCS_BASE_URL + "/config/options_filerenaming_editor.html",
    'doc_cover_art_types':     "https://musicbrainz.org/doc/Cover_Art/Types",
    'plugins':                 "https://picard.musicbrainz.org/plugins/",
    'forum':                   "https://community.metabrainz.org/c/picard",
    'donate':                  "https://metabrainz.org/donate",
    'chromaprint':             "https://acoustid.org/chromaprint#download",
    'acoustid_apikey':         "https://acoustid.org/api-key",
    'acoustid_track':          "https://acoustid.org/track/",
}

# Various Artists MBID
VARIOUS_ARTISTS_ID = '89ad4ac3-39f7-470e-963a-56509c546377'

# Artist alias types
ALIAS_TYPE_ARTIST_NAME_ID = '894afba6-2816-3c24-8072-eadb66bd04bc'
ALIAS_TYPE_LEGAL_NAME_ID = 'd4dcd0c0-b341-3612-a332-c0ce797b25cf'
ALIAS_TYPE_SEARCH_HINT_ID = '1937e404-b981-3cb7-8151-4c86ebfc8d8e'

# Special purpose track titles
SILENCE_TRACK_TITLE = '[silence]'
DATA_TRACK_TITLE = '[data track]'

# Release formats
RELEASE_FORMATS = {}
RELEASE_PRIMARY_GROUPS = {}
RELEASE_SECONDARY_GROUPS = {}
RELEASE_STATUS = {}
for k, v in MB_ATTRIBUTES.items():
    if k.startswith('DB:medium_format/name:'):
        RELEASE_FORMATS[v] = v
    elif k.startswith('DB:release_group_primary_type/name:'):
        RELEASE_PRIMARY_GROUPS[v] = v
    elif k.startswith('DB:release_group_secondary_type/name:'):
        RELEASE_SECONDARY_GROUPS[v] = v
    elif k.startswith('DB:release_status/name:'):
        RELEASE_STATUS[v] = v

# TODO: remove those 4 imports before 3.0 release

# Release countries
from picard.const.countries import (  # noqa: F401,E402 # pylint: disable=unused-import
    RELEASE_COUNTRIES,
)
# List of available user interface languages
from picard.const.languages import (  # noqa: F401,E402 # pylint: disable=unused-import
    UI_LANGUAGES,
)
# List of alias locales
from picard.const.locales import (  # noqa: F401,E402 # pylint: disable=unused-import
    ALIAS_LOCALES,
)
# List of available charsets
from picard.const.scripts import (  # noqa: F401,E402 # pylint: disable=unused-import
    SCRIPTS,
)


# List of official musicbrainz servers - must support SSL for mblogin requests (such as collections).
MUSICBRAINZ_SERVERS = [
    'musicbrainz.org',
    'beta.musicbrainz.org',
]

# Plugins and Release Versions API
PLUGINS_API_BASE_URL = 'https://picard.musicbrainz.org/api/v2/'
PLUGINS_API = {
    'urls': {
        'plugins': PLUGINS_API_BASE_URL + 'plugins/',
        'download': PLUGINS_API_BASE_URL + 'download/',
        'releases': PLUGINS_API_BASE_URL + 'releases',
    },
}

# Default query limit
QUERY_LIMIT = 50

# Maximum number of covers to draw in a stack in CoverArtThumbnail
MAX_COVERS_TO_STACK = 4

# Update levels available for automatic checking
PROGRAM_UPDATE_LEVELS = OrderedDict(
    [
        (
            0, {
                'name': 'stable',
                'title': N_("Stable releases only"),
            }
        ),
        (
            1, {
                'name': 'beta',
                'title': N_("Stable and Beta releases"),
            }
        ),
        (
            2, {
                'name': 'dev',
                'title': N_("Stable, Beta and Dev releases"),
            }
        ),
    ]
)


DEFAULT_PROGRAM_UPDATE_LEVEL = 0


DEFAULT_FILE_NAMING_FORMAT = "$if2(%albumartist%,%artist%)/\n" \
    "$if(%albumartist%,%album%/,)\n" \
    "$if($gt(%totaldiscs%,1),$if($gt(%totaldiscs%,9),$num(%discnumber%,2),%discnumber%)-,)" \
    "$if($and(%albumartist%,%tracknumber%),$num(%tracknumber%,2) ,)" \
    "$if(%_multiartist%,%artist% - ,)" \
    "%title%"


DEFAULT_SCRIPT_NAME = N_("My script")
DEFAULT_COVER_IMAGE_FILENAME = 'cover'
DEFAULT_PROFILE_NAME = N_("My profile")
DEFAULT_COPY_TEXT = N_("(copy)")
DEFAULT_NUMBERED_TITLE_FORMAT = N_("{title} ({count})")
DEFAULT_NAMING_PRESET_ID = "Preset 1"

SCRIPT_LANGUAGE_VERSION = '1.1'

DEFAULT_TIME_FORMAT = '%Y-%m-%d %H:%M:%S'
